#' prepare SWIM gages
#' @description reads SWIM gageinfo dbf as created in gage info pre process and
#' swim points data to create a set of SWIM gages that will be used later.
#' @param SWIM_dbf_file character path tp summary file as created in gageinfo step.
#' @param SWIM_points_file character path to SWIM source data
#' @param ref_gages sf data.frame gages from USGS nldi dataset / ref gages
#' @param gages_add character vector of gages to not filter out. 
#' @param POR_start Date to use as period of record start date
#' @param min_obs minimum number of observations 
#' @return list of three data frames for use later and in mapping output of this 
#' function.
prepare_SWIM <- function(SWIM_dbf_file, SWIM_points_file, ref_gages,
                         gages_add,
                         POR_start = as.Date("1979-10-01"), 
                         min_da_km_gages = 5,
                         min_obs = 90) {
  # Read gagesIII dbf that is a POR summary produced from 00_gage_info.Rmd
  gagesIII_db_full <- read.dbf(file = SWIM_dbf_file, as.is = TRUE) %>%
    rename(total_obs = count_nu, total_poss_obs = TPdays1)
  
  # Read in shapefile
  gagesIII_sf_full <- read_sf(SWIM_points_file, "SWIM_gage_loc") %>%
    st_zm(drop = TRUE) %>%
    # Use site_no as the authoritative station ID name
    rename(site_no = Gage_no) %>%
    # Drop fields in NLDI gages
    select(-c(COMID, REACHCODE, REACH_meas)) %>%
    # Join to NLDI reference gages
    left_join(select(st_drop_geometry(ref_gages), id, comid = nhdpv2_COMID, 
                     reachcode = nhdpv2_REACHCODE, 
                     reach_meas = nhdpv2_REACH_measure), 
              by = c("site_no" = "id")) 
  
  # Read Gages III shapefile
  gagesIII_sf <- gagesIII_sf_full %>% 
    # Get rid of invalid COMIDs, for multiple site IDs get one with the furthest reach measure downstream
    filter(comid > 0, site_no %in% gagesIII_db_full$site_no) %>% 
    # There are some rows with dupslicate site_no, drainage area, and POR infomrmation
    group_by(site_no) %>% 
    slice(which.min(reach_meas)) %>%
    # There were some duplicate rows that needed to be removed
    select(site_no, comid, reachcode, reach_meas) %>% 
    distinct()
  
  # Start documenting gages that are dropped out; these gages have no mean daily Q
  document_dropped_gage(filter(select(st_drop_geometry(gagesIII_sf_full), site_no), 
                               !site_no %in% gagesIII_sf$site_no),
                        "GAGES-III", "No Q or COMID")
  
  # First subset
  gagesIII_db <- gagesIII_db_full %>%
    # Get rid of inital fields we are not interested in, 
    select(-c(agency_cd, ts_id)) %>% 
    #     Constrain by date and/or POR 
    filter(end_date > POR_Start | total_obs > min_obs, 
           #     remove site codes associated with canals, ditchest, etc
           !site_tp_cd %in% c("ST-CA", "ST-DCH", "FA-STS"), 
           #     keep sites with daily mean streamflow parameter
           stat_cd == "00003") %>% 
    distinct() 
  
  # Gages that don't fit the POR requirement, or are non-ST site types
  document_dropped_gage(filter(select(st_drop_geometry(gagesIII_sf), site_no), 
                               !site_no %in% gagesIII_db$site_no),
                        "GAGES-III", "Non-ST; POR")
  
  # scan for potential canals/ditches among the remaining ST site types
  potCanalsIII <- gagesIII_db %>%
    # Flag station_nm with the follwing words
    filter(grepl(paste(c("DITCH", "ditch", "CANAL", "canal", "CN", "DIV", "div"), 
                       collapse = "|"), station_nm)) %>%
    # Remove the station_nm with the following
    filter(!grepl(paste(c("RIVER AT", "RIVER BELOW", "RIVER BL", "R AB", "R BL", "R ABV"), 
                        collapse = "|"), 
                  station_nm, ignore.case = TRUE)) %>%
    # These have been manually checked and are located in gage_mod.R
    filter(!site_no %in% gages_add) 
  
  # Write out for manual checking
  if(!file.exists("temp/GagestoCheck_GAGESIII.csv")){
    write.csv(potCanalsIII, "temp/GagestoCheck_GAGESIII.csv")
  }
  
  # Subset further by streamgages below the minimum drainage area critera
  gagesIII_db_fin <- gagesIII_db %>% 
    filter(!site_no %in% potCanalsIII$site_no, drain_area > (min_da_km_gages * sqkm_sqmi))
  
  # Write out gages that dropped to document.
  document_dropped_gage(select(potCanalsIII, site_no), "GAGES-III", "ST; Pot. Canal")
  document_dropped_gage(gagesIII_db %>%
                          filter(!site_no %in% potCanalsIII$site_no, 
                                 drain_area < min_da_km_gages) %>% 
                          select(site_no),
                        "GAGES-III", "below DA")
  
  # Join shapefile to POR table
  gagesIII_final <- gagesIII_sf %>% 
    inner_join(gagesIII_db_fin, by = "site_no") %>%
    select(-c(coord_acy_, huc_cd, parm_cd, stat_cd, missing_da, site_tp_cd)) %>%
    mutate(source = "gagesIII") %>%
    # deal with one duplicate row
    group_by(site_no) %>%
    filter(total_obs == max(total_obs)) %>%
    ungroup() %>%
    st_transform()
  
  return(list(final = gagesIII_final, full = gagesIII_sf_full, 
              filtered = gagesIII_sf, db = gagesIII_db_full, potCanals = potCanalsIII))
}

#' prepare gage locations from NHDPlusV2
#' @param NHDPlusV2_gageloc_file gageloc info as prepared in 00_gage_info
#' @param NHDPlusV2_gdb character path to national NHDPlus geo database
#' @param SWIM SWIM gage list of data.frames
#' @param ref_gages USGS stream gages from reference gages / NLDI
#' @param min_da_km_gages minimum drainage area for gages to consider
#' @return list of data.frames for later use
prepare_gage_location <- function(NHDPlusV2_gageloc_file, NHDPlusV2_gdb,
                                  SWIM, ref_gages,
                                  min_da_km_gages = 5) {
  # Read gages3 dbf that is a POR summary produced from 00_gage_info.Rmd
  gageloc_db_full <- read.dbf(file = NHDPlusV2_gageloc_file, as.is = TRUE) %>%
    rename(total_obs = count_nu, total_poss_obs = TPdays1)
  
  # Read in shapefile
  gageloc_sf_full <- read_sf(NHDPlusV2_gdb, "Gage") %>%
    # Drop Z geometries
    st_zm(drop = TRUE) %>%
    rename(site_no = SOURCE_FEA) %>%
    select(-c(FLComID, REACHCODE, Measure)) %>%
    # Disregard locations already removed in GAGESIII
    filter(!site_no %in% SWIM$full$site_no) %>%
    # Update index information with NLDI gages
    left_join(select(st_drop_geometry(ref_gages), site_no = id, 
                     comid = nhdpv2_COMID, reachcode = nhdpv2_REACHCODE, 
                     reach_meas = nhdpv2_REACH_measure ), by = "site_no")
  
  # Subset to locations with streamflow
  gageloc_sf <- gageloc_sf_full %>% 
    # Get rid of invalid COMIDs, for multiple site IDs get one with the furthest reach measure downstream
    filter(site_no %in% gageloc_db_full$site_no) %>% 
    # Like GAGES-III, multiple site_no 
    group_by(site_no) %>% 
    slice(which.min(reach_meas)) %>%
    # There were some duplicate rows that needed to be removed
    select(site_no, comid, reachcode, reach_meas) %>% 
    distinct()
  
  # Start documenting gages that are dropped out; these gages have no mean daily Q
  document_dropped_gage(filter(select(st_drop_geometry(gageloc_sf_full), site_no), 
                               !site_no %in% gageloc_sf$site_no), 
                        "GageLoc", "No Q, COMID")
  
  # do the same exercise with GageLoc, but with more site type attributes to filter out
  gageloc_db <- gageloc_db_full %>%
    filter(end_date > POR_Start | total_obs > min_obs, 
           !site_tp_cd %in% 
             c("ST-CA", "ST-DCH", "FA-STS", "SB_TSM", "FA-DV", "FA-STS", "FA-CS",
               "FA-WWD", "FA-OF", "GW"), stat_cd == "00003") %>%
    # Some sites are listed twice with same information but different agencies, 
    select(-c(agency_cd, ts_id))  %>% 
    distinct(site_no, drain_area, dec_lat_va, dec_long_v, .keep_all = T)
  
  # Gages that don't fit the POR requirement, or are non-ST site types
  document_dropped_gage(filter(select(gageloc_db_full, site_no), 
                               !site_no %in% gageloc_db$site_no), 
                        "GageLoc", "Non-ST; POR")
  
  # scan for potential canals/ditches 
  potCanals_gageloc <- gageloc_db %>% 
    filter(grepl(paste(c("DITCH", "ditch", "CANAL", "canal", "CN", "DIV", "div"), 
                       collapse = "|"), station_nm)) %>%
    filter(!site_no %in% gages_add)
  
  # Write out for manual checking
  if(!file.exists("temp/GagestoCheck_GageLoc.csv")){
    write.csv(potCanals_gageloc, "temp/GagestoCheck_GageLoc.csv")
  } else {
    potCanals_gageloc <- read.csv("temp/GagestoCheck_GageLoc.csv")
  }
  
  gageloc_db_fin <- gageloc_db %>% 
    filter(!site_no %in% potCanals_gageloc$site_no, drain_area > 
             (min_da_km_gages * sqkm_sqmi)) 
  
  # Gages that are ST site types, but reside on canals, artificial waterbodies.
  document_dropped_gage(select(potCanals_gageloc, site_no), "GageLoc", "ST; Pot. Canal")
  document_dropped_gage(gageloc_db_fin %>%
                          filter(!site_no %in% potCanals_gageloc$site_no, 
                                 drain_area < min_da_km_gages) %>% select(site_no),
                        "GageLoc", "below DA")
  
  # NHD gages shpafile
  gageloc_final <-  gageloc_sf %>% 
    inner_join(gageloc_db_fin, by = "site_no") %>% 
    select(-c(coord_acy_, huc_cd, parm_cd, stat_cd, missing_da, site_tp_cd)) %>% 
    st_as_sf(coords = c("dec_long_v", "dec_lat_va")) %>%
    mutate(source = "GageLoc") %>%
    # deal with one duplicate row
    group_by(site_no) %>%
    filter(total_obs == max(total_obs)) %>%
    ungroup() %>%
    st_sf() %>%
    st_compatibalize(SWIM$final)
  
  list(final = gageloc_final, filtered = gageloc_sf, full = gageloc_sf_full, 
       db = gageloc_db_full, potCanals = potCanals_gageloc)
}

#' prepare GFv11 gages
#' @param gfv11_gages_rds rds copy of POIs that are gages from gfv11
#' @param gfv11_gages_dbf dbf of gf gages as output from 00_gage_info
#' @param usgs_ref_gages USGS stream gages from reference gages / NLDI
#' @param NHDPlusV2_gages list of data.frames from prepare_gage_locations
#' @param SWIM list of data.frames from prepare_SWIM
#' @param reference_flowlines sf data.frame containing reference flowlines
#' @return list with one "final" data.frame for use in later steps
prepare_gfv11_gages <- function(gfv11_gages_rds, gfv11_gages_dbf, usgs_ref_gages,
                               NHDPlusV2_gages, SWIM, reference_flowlines) {

  # Integrate GFv11 locations not included in GagesLoc and/or gagesIII
  gfv11_sf_full <- readRDS(gfv11_gages_rds) %>%
    st_zm(drop = TRUE) %>%
    st_transform(st_crs(NHDPlusV2_gages$final))
  
  # Filter to sites not in GagesIII or GageLoc
  gfv11_db_full <- read.dbf(file = gfv11_gages_dbf, as.is = TRUE) %>%
    #filter(site_no %in% gfv11_sf_full$site_no) %>%
    select(-c(agency_cd, site_tp_cd, coord_acy_, huc_cd, parm_cd, stat_cd, ts_id, missing_da)) %>%
    rename(total_obs = count_nu, total_poss_obs = TPdays1)
  
  # Remove transboundary
  gfv11_sf <- gfv11_sf_full %>%
    # Remove streamgages assocaited with transboundary
    filter(nchar(NHD_Unit) < 4, 
           !Type_Gage %in% c(SWIM$final$site_no, NHDPlusV2_gages$final$site_no)) %>%
    select(site_no = Type_Gage, NHDPlusID) %>%
    # join to get comid, reachcode, measure data
    left_join(st_drop_geometry(NHDPlusV2_gages$full), by = "site_no") %>%
    # Know from previous work 3 locations with no COM/RC/M
    left_join(select(st_drop_geometry(reference_flowlines), COMID, REACHCODE, FromMeas), 
              by = c("NHDPlusID" = "COMID")) %>%
    # bring over RC/M for 3 gages missing them
    mutate(comid = ifelse(is.na(comid), NHDPlusID, comid),
           reachcode = ifelse(is.na(reachcode), REACHCODE, reachcode),
           reach_meas = ifelse(is.na(reach_meas), FromMeas, reach_meas)) %>%
    select(-c(NHDPlusID, REACHCODE, FromMeas)) %>%
    # Update index information with NLDI gages
    left_join(select(st_drop_geometry(usgs_ref_gages), 
                     id, nhdpv2_COMID, nhdpv2_REACHCODE, nhdpv2_REACH_measure), 
              by = c("site_no" = "id")) %>%
    mutate(comid = ifelse(!is.na(nhdpv2_COMID), nhdpv2_COMID, comid),
           reachcode = ifelse(!is.na(nhdpv2_REACHCODE), 
                              nhdpv2_REACHCODE, reachcode),
           reach_meas = ifelse(!is.na(nhdpv2_REACH_measure), 
                               nhdpv2_REACH_measure, reach_meas)) %>%
    select(site_no, comid, reachcode, reach_meas) %>%
    filter(!is.na(comid)) %>%
    st_sf()
  
  # get NWIS, non-GF lat lon for GFv11 gages
  gfv11_nwis <- dataRetrieval::readNWISsite(gfv11_sf$site_no) %>%
    filter(agency_cd == "USGS") %>%
    select(site_no, dec_long_va, dec_lat_va) %>%
    st_as_sf(coords = c("dec_long_va", "dec_lat_va"), crs = 4269)
  
  # Bring NWIS geometries over to GFv11 data
  gfv11_sf_final <- st_drop_geometry(gfv11_sf) %>%
    inner_join(gfv11_nwis, by = "site_no") %>%
    st_as_sf()
  
  # Build list of STAID - Drainage area for use in the GFv11 section
  # combine DA data from dbf outputs from 00_get_data
  DA_data <- rbind(select(gfv11_db_full, site_no, drain_area), 
                   select(NHDPlusV2_gages$db, site_no, drain_area), 
                   select(SWIM$db, site_no, drain_area)) %>%
    dplyr::filter(!is.na(drain_area)) %>%
    dplyr::rename(drain_area_full = drain_area) %>%
    dplyr::filter(site_no %in% gfv11_sf_final$site_no) %>%
    distinct()
  
  # final Gfv11 file, rebuild geometry
  gfv11_final <- gfv11_sf_final %>%
    left_join(gfv11_db_full, by = "site_no") %>% 
    mutate(source = "GFv11") %>%
    left_join(DA_data, by = "site_no") %>%
    dplyr::mutate(drain_area = if_else(is.na(drain_area), drain_area_full, drain_area)) %>%
    select(-drain_area_full) %>%
    st_compatibalize(SWIM$final)
  
  # Gages that don't fit the POR requirement, or are non-ST site types
  document_dropped_gage(filter(st_drop_geometry(gfv11_sf) %>% select(site_no), !site_no %in% gfv11_db_full$site_no),
                        "GFv11", "No Q")
  
  return(list(final = gfv11_final))
}

#' finalize gage selections
#' @param NHDPlusV2_gages list of data.frames from prepare_gage_locations
#' @param SWIM list of data.frames from prepare_SWIM
#' @param gfv11_gages list with finalized gfv11 gages
#' @param gages_ii_ref data.frame of gages_ii reference gages
#' @param gage_info_gpkg character path to output gages to
#' @param gage_info_csv character path to output gages to
#' @param gage_info_csvt character path to output gages to
#' @param geo_crs coordinate reference system for gage output
finalize_gages <- function(SWIM, NHDPlusV2_gages, gfv11_gages, gages_ii_ref, 
                           gage_info_gpkg, gage_info_csv, gage_info_csvt, 
                           geo_crs = 4326) {
  
  # Bind them together
  gage_final <- data.table::rbindlist(list(SWIM$final, NHDPlusV2_gages$final, 
                                           gfv11_gages$final)) %>%
    st_as_sf()
  
  # which gagesii_ref gages are missing from final set
  missing_gagesii_ref <- filter(gages_ii_ref, !id %in% gage_final$site_no) %>%
    rename(site_no = id, comid = nhdpv2_COMID, reachcode = nhdpv2_REACHCODE, 
           reach_meas = nhdpv2_REACH_measure) %>%
    inner_join(NHDPlusV2_gages$db, by = "site_no") %>%
    select(-c(agency_cd, site_tp_cd, coord_acy_, huc_cd, parm_cd, stat_cd, ts_id, missing_da)) %>%
    st_compatibalize(gage_final)
  
  # Append those with data to gage_final
  gage_final <- data.table::rbindlist(list(gage_final, missing_gagesii_ref), fill = TRUE) %>%
    st_as_sf()
  
  # Rank multiple gages that occur on same COMID
  gage_final_ranked <- gage_final %>% 
    filter(!is.na(comid)) %>%
    group_by(comid) %>%
    mutate(n = n()) %>%
    # Arrange by number of days
    arrange(total_obs, .by_group = TRUE) %>%
    # Populate by rank (1 - lowest number of days to n(highest))
    mutate(gage_score = ifelse(n > 1, row_number(), 1)) %>%
    # If a reference gage is one of gages, bump that to the top
    rowwise() %>%
    mutate(gage_score = ifelse(n > 1,
                               ifelse(!is.na(gagesII_cl) && gagesII_cl == "Ref", 
                                      max(gage_score) + 1, gage_score), gage_score)) %>%
    ungroup() %>%
    # Some gages have multiple gages with same days (inlets/outlets around dam)
    # This further prioritizes to the gage with the smallest reach measure (furthest downstream)
    group_by(comid, total_obs) %>%
    mutate(n = n()) %>%
    arrange(reach_meas, .by_group = TRUE) %>%
    rowwise() %>%
    mutate(gage_score = ifelse(n > 1, 
                               ifelse(reach_meas == min(reach_meas), max(gage_score) + 1, gage_score), gage_score)) %>%
    ungroup() %>%
    group_by(comid) %>%
    mutate(n = n()) %>%
    arrange(end_date, -reach_meas) %>%
    rowwise() %>%
    mutate(gage_score = ifelse(n > 1, 
                               ifelse(end_date == max(end_date), max(gage_score) + 1, gage_score), gage_score)) %>%
    ungroup() %>%
    select(-n) %>%
    st_transform(geo_crs)
  
  # conver to emtpy strings
  write_sf(gage_final_ranked, gage_info_gpkg, "Gages")
  
  # Format csv and csvt file
  # Convert geometry to WKT format
  gage_final_ranked <- st_as_sf(gage_final_ranked, wkt = "geometry")
  
  # Format to WKT geometry field
  gage_final_ranked$WKT <- paste0("POINT", noquote('('), data.frame(st_coordinates(gage_final_ranked))$X, ' ',
                                  data.frame(st_coordinates(gage_final_ranked))$Y, noquote(')'))
  
  # Classes of columns
  gage_final_ranked_csvt <- unlist(sapply(st_drop_geometry(gage_final_ranked), class))
  # replace names
  gage_final_ranked_csvt <- gage_final_ranked_csvt[names(gage_final_ranked_csvt) != "WKT"]
  gage_final_ranked_csvt <- gsub("character", "String", gage_final_ranked_csvt)
  gage_final_ranked_csvt <- gsub("numeric", "Real", gage_final_ranked_csvt)
  gage_final_ranked_csvt <- gsub("Date", "Date (YYYY-MM-DD)", gage_final_ranked_csvt)
  
  # write csv
  write.csv(st_drop_geometry(gage_final_ranked), gage_info_csv, row.names = F, quote = F)
  
  # Write csvt
  csvt_names <- as.matrix(t(unname(gage_final_ranked_csvt)))
  write.table(csvt_names, file = gage_info_csvt, sep = ",", 
              col.names = FALSE, row.names = F)
  
  return(list(final = gage_final_ranked))
}

#' prepare vpu base layers
#' @param reg_gpkg character path to geopackage containing reference fabric for this vpu
#' @param nav_gpkg character path to geopackage where vpu base layers will be written
#' @param vpu vpu that will be written into output list
#' @param full_cat_table table of all catchments for each vpu
#' @param rpu_vpu data.frame ontaining all rpu-vpu pairs
#' @return a list with the vpu and three tables for flowlines, waterbodies, and catchments
prepare_vpu_base_layers <- function(ref_gpkg, nav_gpkg, vpu, full_cat_table, rpu_vpu) {

  fline <- read_sf(ref_gpkg, "flowlines") 
  catchment <- read_sf(ref_gpkg, "catchments") 
  waterbodies <- read_sf(ref_gpkg, "waterbodies")
  
  # we can remove truely degenerate COMIDs 
  # for 0 upstream area and no catchment area
  degen_comid <- fline[fline$TotDASqKM == 0 & 
                         !fline$COMID %in% catchment$featureid, ]$COMID
  
  # need to make sure we don't disconnect anything.
  keep_tocomid <- fline$toCOMID[!fline$COMID %in% degen_comid]
  
  if(length(degen_comid[degen_comid %in% keep_tocomid]) > 0) stop("this will break the network")
  
  fline <- fline[!fline$COMID %in% degen_comid, ]
  
  # there are some duplicates with enhd
  nhd <- fline %>% 
    group_by(COMID) %>%
    filter(row_number() == 1) %>%
    ungroup() %>%
    make_standalone()
  
  # Filter and write dendritic/non-coastal subset to gpkg
  # This will be iterated over to produce the final network after POIs identified
  zero_order <- filter(nhd, TerminalFl == 1 & TotDASqKM < min_da_km_gages)
  non_dend <- unique(unlist(lapply(zero_order$COMID, \(x, nhdDF) get_UM(nhdDF, x), nhdDF = st_drop_geometry(nhd))))
  nhd <- nhd %>%
    mutate(dend = ifelse(!COMID %in% non_dend, 1, 0),
           poi = ifelse(!COMID %in% non_dend & TotDASqKM >= min_da_km_gages, 1, 0)) 
  
  cat_network <- st_drop_geometry(nhd)
  names(cat_network) <- tolower(names(cat_network))
  cat_network <- select(cat_network, comid, tocomid, lengthkm, areasqkm, totdasqkm, 
                        hydroseq, levelpathi, terminalpa, dnlevelpat, dnhydroseq,
                        reachcode, frommeas, tomeas, pathlength, arbolatesu,
                        ftype, fcode, vpuid, rpuid, wbareacomi)
  
  write_sf(nhd, nav_gpkg, nhd_flowline)
  write_sf(cat_network, nav_gpkg, nhd_network)
  write_sf(waterbodies, nav_gpkg, nhd_waterbody)
  
  rpus <- rpu_vpu[rpu_vpu$vpuid == vpu,]$rpuid
  
  cats_rpu <- full_cat_table %>%
    filter(RPUID %in% rpus)
  
  cats <- catchment %>% 
    rename(FEATUREID = featureid, 
           AREASQKM = areasqkm) %>%
    mutate(VPUID = vpu) %>%
    left_join(select(cats_rpu, FEATUREID, RPUID), by = c("FEATUREID")) %>%
    filter(FEATUREID %in% 
             unique(c(nhd$COMID, 
                      full_cat_table$FEATUREID[full_cat_table$RPUID %in% rpus]))) %>%
    mutate(hy_cats = ifelse(FEATUREID %in% nhd$COMID, 1, 0),
           full_cats = ifelse(FEATUREID %in% cats_rpu$FEATUREID, 1, 0)) %>%
    filter(full_cats == 1 | hy_cats == 1) %>%
    st_sf()
  
  write_sf(cats, nav_gpkg, nhd_catchment)

  rpu_list <- setNames(rep(list(list()), length(rpus)), rpus)
  
  for(rpu_code in rpus) {
    
    out_refac_gpkg <- file.path("cache", paste0(rpu_code, "_refactor.gpkg"))
    
    nhd_sub <- subset_rpu(nhd, rpu_code, run_make_standalone = TRUE) %>%
      st_sf()
    
    write_sf(nhd_sub, out_refac_gpkg, nhd_flowline)
    
    cats_rpu <- full_cat_table %>%
      filter(RPUID == rpu_code)
    
    catchments <- cats %>%
      mutate(hy_cats = ifelse(FEATUREID %in% nhd_sub$COMID, 1, 0),
             full_cats = ifelse(FEATUREID %in% cats_rpu$FEATUREID, 1, 0)) %>%
      filter(full_cats == 1 | hy_cats == 1) %>%
      st_sf()
    
    write_sf(catchments, out_refac_gpkg, nhd_catchment)
    
    rpu_list[[rpu_code]] <- list(flowline = nhd_sub, catchment = catchments)
    
  }
  
  return(list(vpu = vpu, flowline = nhd, catchment = cats, catchment_network = cat_network, waterbodies = waterbodies, rpus = rpu_list))
}

#' prepare rpu vpu output list
#' @param vpu_base list of data as output from prepare_vpu_base_layers
#' @param rpu_vpu data.frame ontaining all rpu-vpu pairs
#' @returns list of outputs from a vpu and rpus that make it up
prep_rpu_vpu_out_list <- function(vpu_base, rpu_vpu) {

  nhd <- vpu_base$flowline
  vpu <- vpu_base$vpu
  
  out_vpu <- nhd %>%
    st_drop_geometry() %>%
    select(COMID, toCOMID) %>%
    filter(!toCOMID %in% COMID & !toCOMID == 0)
  
  rpus <- rpu_vpu[rpu_vpu$vpuid == vpu,]$rpuid
  
  rpu_list <- setNames(rep(list(list()), length(rpus)), rpus)
  
  for(rpu_code in rpus) {
    
    nhd_sub <- vpu_base$rpus[[rpu_code]]$flowline
    
    rpu_list[[rpu_code]] <- nhd_sub %>%
      st_drop_geometry() %>%
      select(COMID, toCOMID) %>%
      filter(!toCOMID %in% COMID & !toCOMID == 0)
  }
  return(list(vpu = vpu, vpu_out = out_vpu, rpu_out = rpu_list))
}

#' prepare rpu vpu out
#' @param rpu_vpu_out_list list of rpu vpu outlets
#' @param fline reference flowlines
#' @returns table of flowlines that are rpu and/or vpu outlets
prep_rpu_vpu_out <- function(rpu_vpu_out_list, fline) {

  make_df <- function(x, d, n) {
    y <- d[[x]]
    nr <- nrow(y)
    na <- names(d)[x]
    o <- data.frame(d = rep(na, nr), 
                    COMID = d[[x]]$COMID, 
                    toCOMID = d[[x]]$toCOMID)
    names(o) <- c(n, "COMID", "toCOMID") 
    o
  }
  
  out_rpu <- unlist(lapply(unname(rpu_vpu_out_list), \(x) x$rpu_out), recursive = FALSE)
  out_vpu <- lapply(rpu_vpu_out_list, \(x) x$vpu_out)
  
  names(out_vpu) <- sapply(rpu_vpu_out_list, \(x) x$vpu)
  
  rpu <- do.call(rbind, lapply(1:length(out_rpu), make_df, d = out_rpu, n = "rpu"))
  
  vpu <- do.call(rbind, lapply(1:length(out_vpu), make_df, d = out_vpu, n = "vpu"))
  
  out_rpu_vpu <- left_join(rpu, vpu, by = "COMID") 
  
  out_rpu_vpu <- select(out_rpu_vpu, RPUID = rpu, VPUID = vpu, COMID = COMID, toCOMID = toCOMID.x)
  
  out_rpu_vpu <- left_join(out_rpu_vpu, select(sf::st_drop_geometry(fline), COMID, toRPUID = RPUID, toVPUID = VPUID), 
                           by = c("toCOMID" = "COMID"))
  
  out_rpu_vpu
  
}

#' make combined waterbodies
#' @param nhdplus_gdb NHDPlusV2 Geodatabase path
#' @param proj_crs projected crs to use
#' @return sf data.frame combining the NHDWaterbody and NHDArea layers
#' 
make_combined_waterbodies <- function(nhdplus_gdb, proj_crs) {
  data.table::rbindlist(
    list(
      
      read_sf(nhdplus_gdb, "NHDWaterbody") |>
        st_transform(proj_crs) |>
        mutate(layer = "NHDWaterbody"), 
      
      read_sf(nhdplus_gdb, "NHDArea") |>
        st_transform(proj_crs) |>
        mutate(layer = "NHDArea")
      
    ), fill = TRUE) |>
    st_as_sf()
}
