#' create hu12 POIs
#' @param data hu12_points
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
#' @param vpu vpu that is being executed
#' @returns hu12 poi set
create_hu12_pois <- function(data, flowline, poi_name, vpu) {
  
  # Some NHDPlus VPUs include HUCs from other VPUs
  if(vpu == "02"){
    grep_exp <-"^02|^04"
  } else if (vpu == "08") {
    grep_exp <- "^03|^08"
  } else {
    grep_exp <- paste0("^", substr(vpu, start = 1, stop = 2))
  }
  
  # Join HUC12 outlets with NHD
  HUC12_COMIDs <- data %>% 
    filter(grepl(grep_exp, .data$HUC12)) %>%
    select(COMID, HUC12) %>%
    # Remove this when HUC12 outlets finished
    group_by(COMID) %>% 
    filter(row_number() == 1) %>%
    ungroup()
  
  # Create POIs - some r05 HUC12 POIs not in R05 NHD
  list(POIs = POI_creation(st_drop_geometry(HUC12_COMIDs), filter(flowline, poi == 1), poi_name),
       HUC12_COMIDs = HUC12_COMIDs, vpu = vpu)
  
}

#' create gage pois
#' @param gages as output from 01_prep finalize gages
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
#' @param POIs pois from previous target
#' @param min_da_km_gages minimum drainage area for gages to be included
#' @param combine_meters length to be used later in the workflow to combine flowlines
#' @param reach_meas_thresh a threshold used to determine if a gage will split a flowline or not
#' 
create_gage_pois <- function(gages, flowline, poi_name, POIs,
                             min_da_km_gages = 5, 
                             combine_meters, reach_meas_thresh) {

  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  # Previously identified streamgages within Gage_Selection.Rmd
  streamgages_VPU <- gages %>%
    rename(COMID = comid) %>%
    filter(COMID %in% flowline$COMID) %>%
    inner_join(st_drop_geometry(st_drop_geometry(flowline), COMID, Divergence), by = "COMID") %>%
    switchDiv(., flowline) %>%
    filter(drain_area > min_da_km_gages)
  
  streamgages <- streamgages_VPU %>% 
    group_by(COMID) %>%
    # If multiple gages per COMID, pick one with highest rank as rep POI_ID
    filter(gage_score == max(gage_score), !is.na(drain_area)) %>%
    ungroup() 
  
  # Derive GAGE POIs; use NHD as we've already filtered by NWIS DA in the Gage selection step
  gages_POIs <- gage_POI_creation(POIs, streamgages, filter(flowline, poi == 1), 
                                  combine_meters, reach_meas_thresh, poi_name)
  
  POIs <- gages_POIs$POIs
  
  events <- rename(gages_POIs$events, POI_identifier = Type_Gages)
  
  unassigned_gages <- filter(streamgages_VPU, !site_no %in% POIs$Type_Gages) %>%
    rename(nhd_reachcode = REACHCODE)
  
  if(nrow(unassigned_gages)> 0) {
    # Start documenting gages that are dropped out; these gages have no mean daily Q
    document_dropped_gage(select(unassigned_gages, site_no),
                          "Gages_info", paste0("VPU ", vpu, "; low gage score"))
  }
  
  list(POIs = POIs, events = events, unassigned_gages = unassigned_gages, vpu = vpu)
  
}

#'  Creates POIs for gages using refactor criteria
#'  @param POIs  (sf data.frame) current data frame of POIs
#'  @param gages_info (sf data.frame) VPU specific streamgages from 01_gage_selection
#'  @param nhd (sf data.frame) flowline data.frame 
#'  @param combine_meters (integer) refactor threshold (m) for if if two adjacent fl should be combined
#'  @reach_meas_thresh (integer) threshold added or substracted from reach_meas to determine if a gage will split fl
#' 
#'  @return (sf data.frame) dataframe of gage POIs
#'  
gage_POI_creation <- function(POIs, gages_info, nhd, combine_meters, reach_meas_thresh, poi_name){
  
  # Create streamgage POIs
  gage_POIs <- POI_creation(select(st_drop_geometry(gages_info), COMID, site_no), nhd, poi_name) %>%
    st_compatibalize(., POIs)
  
  # Avoid these for refactoring
  avoid <- dplyr::filter(nhd, (sqrt(AreaSqKM) / LENGTHKM) > 3 & AreaSqKM > 1)
  
  # Create events for streamgages for flowline splitting
  events <- gages_info %>%
    # bring over NHD information
    inner_join(select(st_drop_geometry(gage_POIs), Type_Gages), by = c("site_no" = "Type_Gages")) %>%
    # Get over the NHD attributes
    #inner_join(select(st_drop_geometry(nhd), AreaSqKM, REACHCODE, LENGTHKM, COMID, FromMeas, ToMeas), 
    #           by = "COMID") %>%
    # Apply reach measure thresholds
    filter(reach_meas - FromMeas > reach_meas_thresh & AreaSqKM > 2 & 
             ToMeas - reach_meas > reach_meas_thresh & LENGTHKM > (combine_meters / m_per_km)) %>%
    select(COMID, REACHCODE = reachcode, REACH_meas = reach_meas, Type_Gages = site_no) %>%
    filter(!COMID %in% avoid$COMID) %>%
    mutate(event_type = "streamgage") %>%
    st_compatibalize(., POIs) %>%
    mutate(nexus = TRUE)
  
  # Reset gage flag if even is created
  gage_POIs_nonevent <- filter(gage_POIs, !Type_Gages %in% events$Type_Gages) %>%
    addType(., POIs, poi_name, nexus = FALSE, bind = TRUE) 
  
  if(nrow(events) > 0){
    POIs <- data.table::rbindlist(list(gage_POIs_nonevent, 
                                           select(events, COMID, Type_Gages, nexus)), fill = TRUE) %>%
      mutate(nexus = ifelse(is.na(nexus), FALSE, nexus)) %>%
      st_as_sf()
  } else {
    events <- NA
    POIs <- mutate(POIs, nexus = FALSE)
  }
  
  
  return(list(events = events, POIs = POIs))
}

#' create thermo electric POIs
#' @param te_points point locations of thermo electric plants
#' @param te_attributes attributes for te_points
#' @param POIs pois from previous target
#' @param HUC12_COMIDs HUC12_COMIDs from hu12 poi step
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
#' 
create_thermo_electric_pois <- function(te_points, te_attributes, POIs, HUC12_COMIDs, 
                                        flowline, poi_name) {

  vpu <- POIs$vpu
  POIs <- POIs$POIs

  if(vpu == "08"){
    flowline$VPUID <- "08"
  } else {
    flowline$VPUID <- substr(flowline$RPUID, 1, 2)
  }
  
  # Read in Thermoelectric shapefile for 2015 estimates
  TE_v1 <- te_points
  # Read in Thermoelectric Plants csv for updated estimates  
  TE_v2 <- te_attributes %>%
    mutate(Plant.Code = as.integer(Plant.Code))
  
  # Bind hydrographic address information together and get final COMIDs
  TE_COMIDs_full <- TE_v2 %>%
    filter(water_source %in% c("sw_fresh", "gw_fresh", "gw_sw_mix")) %>%
    left_join(select(st_drop_geometry(TE_v1), 
                     Plant.Code = EIA_PLANT_, COMID), by = "Plant.Code") %>%
    left_join(rename(st_drop_geometry(HUC12_COMIDs), huc12_comid = COMID),
              by = c("huc_12" = "HUC12")) %>%
    mutate(COMID = ifelse(is.na(COMID), huc12_comid, COMID)) %>%
    select(-huc12_comid)
  
  # Prepare TE for POI Creation
  TE_COMIDs <- TE_COMIDs_full %>%
    inner_join(select(st_drop_geometry(flowline), COMID, VPUID), by = "COMID") %>%
    filter(grepl(paste0("^", substr(vpu, 1, 2), ".*"), .data$VPUID), COMID > 0) %>%
    switchDiv(., flowline) %>%
    group_by(COMID) %>%
    summarize(eia_id = paste0(unique(Plant.Code), collapse = " "), count = n()) %>%
    ungroup()
  
  unassigned_plants <- filter(TE_COMIDs, !COMID %in% POIs$COMID)
  
  if(nrow(TE_COMIDs) > 0){
    # Derive TE POIs
    POIs <- POI_creation(st_drop_geometry(TE_COMIDs), filter(flowline, poi == 1), poi_name) %>%
      addType(., POIs, "TE", nexus = TRUE) 
     
  }
  
  list(POIs = POIs, unassigned_plants = unassigned_plants, vpu = vpu)
}

#' make waterbodies layer
#' @param combined_waterbodies nhdplus area and waterbody layers combined together
#' @param 
make_waterbodies_layer <- function(combined_waterbodies, flowline, geo_crs = 4326) {
  
  # Waterbodies sourced from NHD waterbody layer for entire VPU
  WBs_VPU_all <- filter(combined_waterbodies, 
                        COMID %in% flowline$WBAREACOMI) %>%
    filter(FTYPE != "SwampMarsh") %>%
    mutate(FTYPE = as.character(FTYPE),
           source = "NHDv2WB",
           wb_id = ifelse(GNIS_ID == " ", COMID, GNIS_ID)) %>%
    st_transform(geo_crs) %>%
    st_make_valid()
  
  old_sf_use_s2 <- sf_use_s2(FALSE)
  
  ref_WB <- WBs_VPU_all %>%
    group_by(wb_id, GNIS_NAME) %>%
    sfheaders::sf_remove_holes() %>%
    summarize(do_union = TRUE, member_comid = paste(COMID, collapse = ",")) %>%
    st_make_valid() %>%
    ungroup() %>%
    mutate(area_sqkm = as.numeric(st_area(.)) / 1000000) %>%
    mutate(source = "NHDv2WB") %>%
    st_cast("MULTIPOLYGON")
  
  sf_use_s2(old_sf_use_s2)
  
  list(wbs_draft = WBs_VPU_all, WBs_layer_orig = ref_WB)
}

#' create resops POIs
#' @param POIs POIs as created by previous step
#' @param wbs waterbodies table
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
#' 
create_resops_pois <- function(POIs, wbs, istarf_xwalk, flowline, poi_name = "resops") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs

  # Unnest wb_poi_lst
  wb_table <- st_drop_geometry(wbs$WBs_layer_orig) %>%
    dplyr::mutate(member_comid = strsplit(member_comid, ",")) %>%
    tidyr::unnest(cols = member_comid) %>%
    mutate(member_comid = as.character(member_comid)) %>%
    distinct()
  
  # ResOpsUS locations with attributes from the VPU NHDv2 wb set
  resops_wb_df <- istarf_xwalk %>%
    # Subset to VPU, only one DAMID per waterbody
    filter(flowlcomid %in% flowline$COMID | 
             wbareacomi %in% wb_table$member_comid) %>%
    dplyr::select(grand_id, nid_source_featureid, source = comi_srclyr,  
                  flowlcomid, wbareacomi, hr_permid, onoffnet) %>%
    mutate(wbareacomi = as.character(wbareacomi)) %>%
    # Link to waterbody table
    left_join(distinct(wb_table, member_comid, wb_id), 
              by = c("wbareacomi" = "member_comid")) %>%
    left_join(select(st_drop_geometry(wbs$wbs_draft), wbareacomi = COMID, 
                     wb_id2 = wb_id) %>%
                mutate(wbareacomi = as.character(wbareacomi)), 
              by = "wbareacomi") %>%
    mutate(wb_id = ifelse(is.na(wb_id), wb_id2, wb_id),
           wbareacomi = ifelse(wbareacomi == -2, 
                               hr_permid, wbareacomi)) %>%
    select(-c(wb_id2)) 
  
  # POIs with waterbody IDs in reference waterbodies
  resops_wb_pois <- filter(wbs$WBs_layer_orig, wb_id %in% resops_wb_df$wb_id) %>%
    inner_join(select(st_drop_geometry(resops_wb_df), grand_id,
                      NID_ID = nid_source_featureid,
                      resops_flowlcomid = flowlcomid, wb_id),
               by = "wb_id") %>%
    #mutate(source = "ref_WB") %>%
    group_by(wb_id) %>%
    filter(n() == 1) %>%
    st_as_sf()
  
  # Add ResOPsUS locations to waterbody list with attributed NID and resops data
  if(nrow(resops_wb_pois) > 0){
    wb_poi_lst_filtered <- filter(wbs$WBs_layer_orig, !wb_id %in% resops_wb_pois$wb_id, 
                                  !is.na(wb_id)) 
    
    wb_poi_lst <- data.table::rbindlist(list(wb_poi_lst_filtered, resops_wb_pois), 
                                        fill = TRUE) %>%
      mutate(accounted = 0) %>%
      st_as_sf()
    
  } else {
    
    wb_poi_lst <- mutate(wbs$WBs_layer_orig, accounted = 0)
    
  }
  
  # Reach resopsus
  reach_resops <- filter(resops_wb_df, !wb_id %in% wb_poi_lst$wb_id,
                         !source %in% c("NHDAREA", "HR ID AVAILABLE") |
                           onoffnet == 0) %>%
    mutate(Type_WBOut = NA) %>%
    select(COMID = flowlcomid, resops = grand_id, Type_WBOut)
  
  # Existing reservoir-waterbody outlet POIs
  exist_POIs_WB <- filter(resops_wb_df, flowlcomid %in% POIs$COMID) %>%
    mutate(wb_id = ifelse(source == "NHDAREA", wbareacomi, wb_id)) %>%
    filter(!is.na(wb_id)) %>%
    select(COMID = flowlcomid, resops = grand_id, Type_WBOut = wb_id)
  
  # Resops POIs
  resops_pois <- rbind(reach_resops, exist_POIs_WB)
  
  # Resops defined by reach
  if(nrow(resops_pois) > 0){
    # Resops POIs with no reservoirs, defined by reach
    POIs <- POI_creation(resops_pois, filter(flowline, poi == 1), poi_name) %>%
      addType(., POIs, poi_name, nexus = TRUE)
    
    # TODO: investigate many to many join
    # Add waterbody attribute
    POIs <- POIs %>%
      left_join(select(resops_pois, COMID, Type_WBOut),
                by = "COMID") %>%
      mutate(Type_WBOut = ifelse(!nexus, 
                                 Type_WBOut, NA))
    
    # Resops with reservoirs
    resops_wb_df <- resops_wb_df %>%
      mutate(accounted = ifelse(grand_id %in% POIs$Type_resops, 1, 0),
             source = ifelse(grand_id %in% reach_resops$resops, "REACH", source))
  }
  list(wb_pois = wb_poi_lst, wb_resops = resops_wb_df, POIs = POIs, vpu = vpu)
  
}

#' create hilarri pois
#' @param POIs POIs as created by previous step
#' @param wbs waterbodies table
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
create_hilarri_pois <- function(POIs, hilarri_data, wb_poi_lst, flowline, poi_name = "hilarri") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  old_sf_use_s2 <- sf::sf_use_s2(FALSE)
  on.exit(sf::sf_use_s2(old_sf_use_s2))
  # 1: Many based on original GNIS_ID
  wb_table <- st_drop_geometry(wb_poi_lst) %>%
    dplyr::mutate(member_comid = strsplit(member_comid, ",")) %>%
    tidyr::unnest(cols = member_comid) %>%
    mutate(member_comid = as.integer(member_comid)) %>%
    distinct()
  
  # Hilarri POI Points
  hilarri_points <- hilarri_data %>%
    mutate(nhdwbcomid = as.integer(nhdwbcomid)) %>%
    filter(!dataset %in% c('Power plant only; no reservoir or inventoried dam'),
           nhdv2comid %in% flowline$COMID) %>%
    left_join(select(wb_table, member_comid, wb_id), 
              by = c("nhdwbcomid" = "member_comid"))
  
  # Waterbodies linked to hilarri information
  hil_wb_pois_rf <- wb_poi_lst %>%
    inner_join(select(st_drop_geometry(hilarri_points), hilarriid, nhdwbcomid, 
                      nidid_hill = nidid, wb_id),
               by = "wb_id") %>%
    group_by(wb_id) %>%
    mutate(hilarriid = paste0(hilarriid, collapse = ","),
           nidid_hill = paste0(nidid_hill, collapse = ",")) %>%
    ungroup() %>%
    distinct() %>%
    st_as_sf() %>%
    st_compatibalize(wb_poi_lst)
  
  # Add ResOPsUS locations to waterbody list 
  if(nrow(hil_wb_pois_rf) > 0){
    wb_poi_lst <- wb_poi_lst %>%
      #filter(!is.na(resops_FlowLcomid)) %>%
      left_join(select(st_drop_geometry(hil_wb_pois_rf), wb_id, hilarriid, 
                       nidid_hill), by = "wb_id") %>%
      mutate(NID_ID = ifelse(is.na(NID_ID), nidid_hill, NID_ID)) %>%
      select(-nidid_hill) %>%
      distinct()
  }
  
  # Reach POIs
  reach_pois <- filter(hilarri_points, !hilarriid %in% wb_poi_lst$hilarriid) %>%
    select(COMID = nhdv2comid, hilarriid) %>%
    mutate(COMID = as.integer(COMID)) %>%
    group_by(COMID) %>%
    filter(row_number() == 1) %>%
    st_drop_geometry()
  
  # Make POIs for reach POIs
  POIs <- POI_creation(reach_pois, filter(flowline, poi == 1), poi_name) %>%
    addType(., POIs, poi_name, nexus = TRUE)
  
  list(POIs = POIs, wb_pois = wb_poi_lst, vpu = vpu)
  
}

#' create waterbody outlet pois
#' @param POIs POIs from previous step
#' @param NHDPlus_NHDArea NHDArea features from NHDPlusV2
#' @param wb_poi_lst wb_pois from hilarri output list
#' @param resops_wb_df wb_resops from resops output list
#' @param flowline flowlines to attach pois to
#' @param events events from gage output list
#' @param proj_crs projected crs for overall project
#' @param gages gages from gage_info_gpkg
#' @param poi_name name for poi set
create_wb_outlet_pois <- function(POIs, 
                                  NHDPlus_NHDArea, 
                                  wb_poi_lst, 
                                  resops_wb_df,
                                  flowline,
                                  events,
                                  proj_crs,
                                  gages, 
                                  poi_name,
                                  nhdplushr_dir) {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs

  # Hillari IDs
  hilarri_pois <- filter(POIs, !is.na(Type_hilarri))
  
  # Get NHDARea polygons already define
  nhd_area_vpu <- NHDPlus_NHDArea %>%
    select(COMID, AREASQKM) %>%
    mutate(source = "NHDv2Area") %>%
    st_compatibalize(wb_poi_lst)
  
  # Non-nhdv2 waterbodies from resopsus
  other_wbs <- filter(resops_wb_df, !wb_id %in% wb_poi_lst$wb_id, 
                      source != "REACH") %>%
    select(grand_id, member_comid = wbareacomi, resops_flowlcomid = flowlcomid, 
           source, accounted) %>%
    left_join(select(st_drop_geometry(hilarri_pois), 
                     COMID, hilarriid = Type_hilarri),
              by = c("resops_flowlcomid" = "COMID")) %>%
    group_by(resops_flowlcomid) %>%
    filter(row_number() == 1) %>%
    ungroup() %>%
    left_join(select(nhd_area_vpu, COMID) %>%
                mutate(COMID = as.character(COMID)),
              by = c("member_comid" = "COMID")) %>%
    st_as_sf() %>%
    mutate(wb_id = ifelse(source == "NHDAREA", member_comid, NA)) %>%
    st_drop_geometry()
  
  # copy empty geometries to wb_list for use in wb POI function
  if(nrow(other_wbs) > 0){
    wb_poi_list_fn <- data.table::rbindlist(list(wb_poi_lst, 
                                                 other_wbs),
                                            fill = TRUE) %>%
      distinct() %>%
      st_as_sf()
  } else {
    wb_poi_list_fn <- wb_poi_lst
  }
  
  # Final waterbody list with all requirements
  final_wb_list <- filter(wb_poi_list_fn,
                          area_sqkm > wb_area_thresh |
                            !is.na(resops_flowlcomid) |
                            !is.na(hilarriid))
  
  new_path <- "?prefix=StagedProducts/Hydrography/NHDPlusHR/VPU/Current/GDB/"
  assign("nhdhr_file_list", new_path, envir = nhdplusTools:::nhdplusTools_env)
  
  # fix something here
  wb_layers <- wbout_POI_creaton(filter(flowline, poi == 1), final_wb_list, 
                                 NHDPlus_NHDArea, POIs, proj_crs, wb_poi_lst, poi_name, nhdplushr_dir)
  
  wb_events <- poi_missing_wb <- wb_missing_poi <- double_wb <- double_wb_pois <- NULL
  
  if(!is.data.frame(wb_layers)){
    # write_sf(wb_layers$nhd_wb_table, temp_gpkg, "wb_flowpaths")
    # 
    wbs <- wb_layers$WBs
    wb_pois <- wb_layers$POIs %>% filter(!is.na(Type_WBOut))
    missing_wb <- filter(wbs, !wb_id %in% wb_pois$Type_WBOut)
    print(nrow(missing_wb))
    
    double_wb <- wbs %>% group_by(wb_id) %>% filter(n() > 1)
    
    double_wb_pois <- wb_pois %>%
      group_by(Type_WBOut) %>%
      filter(n() > 1)
    
    wb_missing_poi <- filter(wbs, 
                             !wb_id %in% wb_pois$Type_WBOut)
    
    poi_missing_wb <- filter(wb_pois, !Type_WBOut %in% wb_layers$WBs$wb_id)
    
    #*************************************
    if(all(!is.na(wb_layers$events))){
      wb_events <- wb_layers$events %>%
        select(COMID, REACHCODE, REACH_meas, POI_identifier = wb_id,
               event_type, nexus) %>%
        st_compatibalize(., events)
      
      events <- rbind(events, wb_events)
    }
    
    POIs <- wb_layers$POIs
    # Write out waterbodies
    WB_VPU_pois <- wb_layers$WBs
  } else {
    POIs <- wb_layers
    WB_VPU_pois <- final_wb_list
  }
  
  #12047928
  wb_out_col <- wb_poi_collapse(POIs, flowline, events, gages)
  POIs <- wb_out_col$POIs
  
  if(all(!is.na(wb_out_col$events_ret))){
    events <- wb_out_col$events_ret
  }
  
  list(POIs = POIs, wb_out_col = wb_out_col, 
       all_events = events, wb_events = wb_events, 
       double_wb = double_wb, double_wb_pois = double_wb_pois, 
       wb_missing_poi = wb_missing_poi, poi_missing_wb = poi_missing_wb, 
       updated_flowline = wb_layers$nhd_wb_table, WB_VPU_pois = WB_VPU_pois, vpu = vpu)
}

#'  Creates Waterbody POIs, calls a few other functions
#'  @param nhd (sf data.frame) flowline data.frame 
#'  @param wb_table (sf data.frame) 
#'  @param nhdplus_area (sf data.frame) NHDPlus Area features
#'  @param crs (integer) CRS to use (epsg code) 
#'  @param wb_poi_lst TODO
#'  @param poi_name name for pois to be created
#'  @param nhdplushr_dir directory containing NHDPlusHR data
#' 
#'  @return (sf data.frame) dataframe of waterbody outlet POIs
#'  
wbout_POI_creaton <- function(nhd, wb_table, nhdplus_area, POIs, crs, 
                              wb_poi_lst, poi_name, nhdplushr_dir){
  
  wb_components_table <- st_drop_geometry(wb_table) %>%
    dplyr::mutate(member_comid = strsplit(member_comid, ",")) %>%
    tidyr::unnest(cols = member_comid) %>%
    mutate(member_comid = as.integer(member_comid)) %>%
    distinct() 
  
  # R17, some wierd stuff going on in the waterbodies
  if("947070203" %in% wb_components_table$member_comid){
    wb_components_table <- wb_components_table %>%
      filter(!wb_id %in% c(1503847, 1513298))
  }
  
  # Bring wb_id to NHD data frame
  if(("wb_id") %in% colnames(nhd)){
    nhd <- select(nhd, -wb_id)
  }
  
  nhd <- nhd %>%
    left_join(distinct(st_drop_geometry(wb_components_table), wb_id, member_comid) %>%
                mutate(member_comid = as.numeric(member_comid)),
              by = c("WBAREACOMI" = "member_comid"))
  
  
  # Resops that are either NHDArea or MHD HR source and are missing in the
  #       nhd attributiong table
  resops_wb_other_sources <- filter(wb_table, source ==
                                      "HR ID AVAILABLE" | (source == "NHDAREA" &
                                                             !wb_id %in% nhd$wb_id) &
                                      accounted == 0)
  
  # Resops that are not accounted for and NHDv2 waterbody or NHDArea sourced
  resops_main <- filter(wb_table, !is.na(resops_flowlcomid) &
                          !wb_id %in% resops_wb_other_sources$wb_id,
                        accounted == 0)
  
  # Remaining WB that don't have outlet COMID defined and aren't in ResOPsUS
  WB_outlet <- filter(wb_table, is.na(resops_flowlcomid),
                      !wb_id %in% POIs$Type_WBOut)
  
  # Size-based waterbody outlet POIs ot in NHDArea, HR, or
  #        manually defined with ResOPS
  wbout_COMIDs <- nhd %>%
    filter(wb_id %in% WB_outlet$wb_id) %>%
    group_by(wb_id) %>%
    slice(which.min(Hydroseq)) %>%
    ungroup() %>%
    switchDiv(., nhd) %>%
    select(COMID, wb_id) %>%
    mutate(wbtype = "size") %>%
    st_drop_geometry() %>%
    distinct()
  
  # ResOpsUS defined outlets
  resops_main_outlets <- select(st_drop_geometry(resops_main), COMID = resops_flowlcomid,
                                wb_id) %>%
    mutate(wbtype = "resops")
  
  #  combine together
  wbout_coms <- distinct(rbind(wbout_COMIDs, resops_main_outlets))
  
  # Get NHDArea and HI-Res waterbodies for ResOpsUS locaitons
  WBs_VPU_areaHR <- HR_Area_coms(nhd, resops_wb_other_sources, nhdplus_area, crs, nhdplushr_dir)
  
  # if there are no HR/Area waterbodies generated
  if(all(is.na(WBs_VPU_areaHR))){
    POIs <- POI_creation(wbout_coms,
                             nhd, "wb_id") %>%
      addType(., POIs, "wb_id", nexus = TRUE) %>%
      mutate(Type_WBOut = ifelse(is.na(Type_WBOut), Type_wb_id, Type_WBOut)) %>%
      select(-Type_wb_id) %>%
      left_join(select(st_drop_geometry(resops_main), grand_id, resops_flowlcomid),
                by = c("COMID" = "resops_flowlcomid")) %>%
      mutate(Type_resops = ifelse(!is.na(grand_id), grand_id, Type_resops)) %>%
      select(-grand_id)
  } else {
    # WBarea out comids
    wbareaout_coms <- select(st_drop_geometry(nhd), COMID, Hydroseq) %>%
      inner_join(filter(WBs_VPU_areaHR$nhd_wb_table, !is.na(outlet)),
                 by = c("COMID" = "comid")) %>%
      group_by(wb_id) %>%
      filter(Hydroseq == min(Hydroseq)) %>% #, row_number() == 1
      ungroup() %>%
      inner_join(select(WBs_VPU_areaHR$wb, -source), by = "wb_id") %>%
      filter(source != "NHDHR") %>%
      select(COMID, wb_id) %>%
      mutate(wbtype = "resops")
    
    if(nrow(wbareaout_coms) > 0){
      wbout_coms <- rbind(wbout_coms, wbareaout_coms)
    }
    
    POIs <- POI_creation(select(wbout_coms, COMID, wb_id),
                             nhd, poi_name) %>%
      addType(., POIs, poi_name, nexus = TRUE) %>%
      mutate(Type_WBOut = ifelse(is.na(Type_WBOut) & (!nexus | is.na(nexus)),
                                 Type_wb_id, Type_WBOut)) %>%
      select(-Type_wb_id) %>%
      left_join(select(st_drop_geometry(resops_main), grand_id, resops_flowlcomid),
                by = c("COMID" = "resops_flowlcomid")) %>%
      mutate(Type_resops = ifelse(!is.na(grand_id) & (!nexus | is.na(nexus)),
                                  grand_id, Type_resops)) %>%
      select(-grand_id)
  }
  
  # If the return is not NA
  if(inherits(WBs_VPU_areaHR, "list")){
    # Attribute flowlines that intersect NHDARea and/or NHD HR waterbodies with
    #           waterbody COMIDs
    # For NHDHR, two waterbodies might be back to back and share an intersecting
    #     flowline, this may cause problems below, so deal with this case
    WBs_VPU_areaHR$nhd_wb_table <- WBs_VPU_areaHR$nhd_wb_table %>%
      left_join(select(st_drop_geometry(resops_wb_other_sources),
                       resops_flowlcomid, member_comid, grand_id),
                by = c("comid" = "resops_flowlcomid",
                       "wb_comid" = "member_comid")) %>%
      group_by(comid) %>%
      filter(case_when(n() == 2 ~ !is.na(grand_id),
                       TRUE ~ n() == 1)) %>%
      ungroup()
    
    nhd <- nhd %>%
      left_join(distinct(WBs_VPU_areaHR$nhd_wb_table, comid, wb_id_hr = wb_id),
                by = c("COMID" = "comid")) %>%
      mutate(wb_id = ifelse(!is.na(wb_id_hr), wb_id_hr, wb_id)) %>%
      select(-wb_id_hr)
    
    # NHD flowlines within the HR waterbodies
    nhd_WBs <- filter(nhd, wb_id %in%
                        filter(WBs_VPU_areaHR$nhd_wb_table, source == "NHDHR")$wb_id)
    
    if(nrow(nhd_WBs) > 0){
      # Create outlet events for splitting HR and NHDArea waterbodies
      wb_outlet_events <- WB_event(WBs_VPU_areaHR, nhd_WBs, "outlet") %>%
        st_compatibalize(., POIs) %>%
        mutate(nexus = TRUE, Type_resops = 2)
      
      POIs <- data.table::rbindlist(list(POIs,
                                         select(wb_outlet_events, COMID, Type_WBOut = wb_id, Type_resops, nexus)), fill = TRUE) %>%
        st_as_sf()
    } else {
      wb_outlet_events <- NA
    }
    
    WBs_VPU_areaHR$wb <- WBs_VPU_areaHR$wb %>%
      sfheaders::sf_remove_holes(.) %>%
      st_compatibalize(wb_poi_lst)
    
    WBs_VPU_areaHR$wb <- nhdplusTools::st_compatibalize(
      WBs_VPU_areaHR$wb, wb_table)
    
    wb_table_filtered <- wb_table %>%
      filter(!st_is_empty(wb_table))
    
    # Append NHDArea and NHDHR waterbodies to waterbody layer
    WBs_VPU <- data.table::rbindlist(list(wb_table_filtered, WBs_VPU_areaHR$wb),
                                     fill = TRUE) %>%
      st_as_sf()
    
    st_geometry(WBs_VPU) <- st_make_valid(st_geometry(WBs_VPU))
    
    return(list(POIs = POIs, events = wb_outlet_events, WBs = WBs_VPU,
                nhd_wb_table = nhd))
    
  } else {
    return(list(POIs = POIs, events = NA, WBs = wb_table,
                nhd_wb_table = nhd))
  }
}

#'  Creates Waterbody POIs, calls a few other functions
#'  @param nhd (sf data.frame) flowline data.frame 
#'  @param wb_lyr (sf data.frame) waterbodies to create inlet pois from
#'  @param events (sf data.frame) of events that waterbody inlets can be pulled from
#'  @param nhdplus_area (sf data.frame) NHDPlus Area features
#'  @param crs (integer) CRS to use (epsg code)
#'  @param POIs (sf data.frame) rolling POI data frame
#'  @param nhdplushr_dir directory containing NHDPlusHR data
#' 
#'  @return (sf data.frame) dataframe of WB inlet POIs
#'  
wbin_POIcreation <- function(flowline, wb_lyr, events, nhdplus_area, crs, 
                             POIs, poi_name = "WBIn", nhdplushr_dir){
  
  wbout_COMIDs <- filter(POIs, !is.na(Type_WBOut))
  WBs_in_POIs <- filter(wb_lyr, wb_id %in% wbout_COMIDs$Type_WBOut)
  
  # Unnest to get list of waterbody COMIDS (member_comid)
  wb_table <- st_drop_geometry(wb_lyr) %>%
    dplyr::mutate(member_comid = strsplit(member_comid, ",")) %>%
    tidyr::unnest(cols = member_comid) %>%
    mutate(member_comid = as.integer(member_comid)) %>%
    distinct()
  
  # # Attribute flowline if waterbody exists for POIs
  # nhd_wb <- nhd %>%
  #   left_join(distinct(wb_table, wb_id, member_comid), 
  #              by = c("WBAREACOMI" = "member_comid")) 
  
  # Create waterbody inlet POIs
  wbin_COMIDs <- filter(flowline, is.na(wb_id) & 
                          DnHydroseq %in% filter(flowline, !is.na(wb_id))$Hydroseq) %>%
    select(-wb_id) %>%
    switchDiv(., flowline) %>%
    inner_join(st_drop_geometry(filter(flowline, minNet == 1)) %>%
                 select(wb_id, Hydroseq), by = c("DnHydroseq" = "Hydroseq")) %>%
    select(COMID, wb_id, minNet) %>%
    group_by(COMID) %>%
    # Ensure the inlets are on the network defined by confluence POIs
    filter(minNet == 1) 
  
  # Headwater Waterbodies that may need the network extended through the inlet
  need_wbin <- st_drop_geometry(wb_lyr) %>%
    #dplyr::select(COMID)%>%
    dplyr::filter(!wb_id %in% wbin_COMIDs$wb_id)
  
  if(nrow(need_wbin)>0){
    
    nhd_inlet <- mutate(flowline, WB = ifelse(wb_id > 0 & wb_id %in% need_wbin$wb_id, 1, 0))
    
    missing_wbin_COMIDs <- filter(nhd_inlet, WB == 0,
                                  DnHydroseq %in% filter(nhd_inlet, WB == 1)$Hydroseq) %>%
      select(-wb_id) %>%
      switchDiv(., nhd_inlet) %>%
      inner_join(st_drop_geometry(filter(nhd_inlet, minNet == 1)) %>%
                   select(Hydroseq, wb_id), 
                 by = c("DnHydroseq" = "Hydroseq")) %>%
      select(COMID, wb_id) %>%
      group_by(COMID) %>%
      filter(row_number() == 1) %>%
      ungroup()
    
    missing_wbin_COMIDs <- missing_wbin_COMIDs %>%
      left_join(select(st_drop_geometry(nhd_inlet), COMID, TotDASqKM)
                , by = c("COMID")) %>%
      group_by(COMID) %>%
      slice(which.max(TotDASqKM))%>%
      ungroup() %>%
      select(-TotDASqKM)
    
    if(nrow(missing_wbin_COMIDs) > 0){
      wbin_COMIDs <- data.table::rbindlist(list(wbin_COMIDs, 
                                                missing_wbin_COMIDs),
                                           fill = TRUE) %>%
        select(COMID, wb_id)
    }
  }
  
  wbin_POIs <- POI_creation(filter(st_drop_geometry(wbin_COMIDs), !COMID %in% wbout_COMIDs$COMID), 
                            flowline, "WBIn")
  
  # Get NHDArea and HR waterbodies
  WBs_VPU_areaHR <- HR_Area_coms(flowline, 
                                 filter(wb_lyr, source == "HR ID AVAILABLE"), 
                                 nhdplus_area, crs, nhdplushr_dir)
  
  if(is.list(WBs_VPU_areaHR)) {
    nhd_WBs <- filter(flowline, minNet == 1,
                      COMID %in% WBs_VPU_areaHR$nhd_wb_table$comid,
                      AreaSqKM > 0)
  } else {
    nhd_WBs <- NA
  }
  
  if(is.data.frame(nhd_WBs)){
    
    # Get the outlet events again
    wb_outlet_events <- events %>%
      filter(event_type == "outlet")
    
    # Drive the inlet events
    nhd_inlet <- filter(nhd_WBs, !COMID %in% wb_outlet_events$COMID)
    
    if("12" %in% unique(flowline$VPUID)){
      nhd_inlet <- rbind(nhd_inlet, filter(nhd_WBs, COMID == 1304709))
    }
    
    if (nrow(nhd_inlet) > 0){
      # Get inlet events and bind with outlets
      wb_inlet_events <- WB_event(WBs_VPU_areaHR, nhd_inlet, "inlet") %>%
        mutate(nexus = TRUE) %>%
        inner_join(select(st_drop_geometry(flowline), COMID, Hydroseq, ToMeas), by = "COMID") 
      
      # Determine which events are close enough to the end of an upstream flowline to just use that geometry for the POI
      wbin_fl_upstream <- filter(flowline, DnHydroseq %in% filter(flowline, COMID %in% wb_inlet_events$COMID)$Hydroseq) %>%
        group_by(DnHydroseq) %>%
        filter(n() == 1) %>%
        ungroup()
      
      if(nrow(wbin_fl_upstream) > 0){
        
        wb_inlet_POIs <- filter(wb_inlet_events, REACH_meas < 5 | as.integer(REACH_meas) == as.integer(ToMeas), 
                                COMID %in% wbin_fl_upstream$toCOMID) %>%
          inner_join(select(st_drop_geometry(wbin_fl_upstream), usCOMID = COMID, toCOMID), by = c("COMID" = "toCOMID")) %>%
          mutate(dsCOMID = COMID, COMID = usCOMID)
        
        if(nrow(wb_inlet_POIs) > 0) {
          wbin_POIs <- bind_rows(wbin_POIs, POI_creation(select(st_drop_geometry(wb_inlet_POIs), dsCOMID, 
                                                                Type_WBIn = POI_identifier), 
                                                         flowline, "WBIn"))
          
          wb_inlet_events <- filter(wb_inlet_events, !COMID %in% wb_inlet_POIs$dsCOMID)
        }
      }
      
      wbin_POIs <- addType(wbin_POIs, POIs, "WBIn", nexus = TRUE)
      
      if(nrow(wb_inlet_events) > 0){
        
        wb_inlet_events <- st_compatibalize(wb_inlet_events, wbin_POIs)
        wbin_POIs <- data.table::rbindlist(list(wbin_POIs, 
                                                select(wb_inlet_events, COMID, 
                                                       Type_WBIn = POI_identifier, nexus)), 
                                           fill = TRUE) %>%
          st_as_sf()
      }
      
      return(list(POIs = wbin_POIs, events = wb_inlet_events))
    } else {
      print("no waterbody inlets events")
      wbin_POIs <- addType(wbin_POIs, POIs, "WBIn")
      return(list(POIs = wbin_POIs, events = NA))
    }
  } else {
    wbin_POIs <- addType(wbin_POIs, POIs, "WBIn")
    return(list(POIs = wbin_POIs, events = NA))
  }
  
}

### USEDIN wbout_POI_creaton and wbin_POIcreation
#'  Retrieves waterbodies from NHDArea layer or NHD Hi-Res for ResOpsUs 
#'  locations if absent from NHD waterbody
#'  @param nhd (sf data.frame) Fowlines for given VPU
#'  @param WBs  (sf data.frame) waterbodiess for discretization within VPU
#'  @param nhdplus_area (sf data.frame) NHDPlus Area features
#' @param crs coordinate reference system to use for analysis
#' @param nhdplushr_dir directory containing NHDPlusHR data
#' 
#'  @return (list) wbs - sf data frame for NHDArea and HR waterbodies
#'                 wb_table - table of flowlines and outlet info for each 
#'                            feature in wb
HR_Area_coms <- function(nhd, WBs, nhdplus_area, crs, nhdplushr_dir){
  
  # Pull out rows for VPU that are NHDArea
  nhd_area_resops <- WBs %>%
    filter(resops_flowlcomid %in% nhd$COMID, source == "NHDAREA")
  
  # Pull out rows for VPU that are NHD HR
  nhd_hr_wb_resops <-  WBs %>%
    filter(member_comid != 65000300139130) %>% # R09, in Canada
    filter(resops_flowlcomid %in% nhd$COMID, source == "HR ID AVAILABLE") %>%
    st_drop_geometry()
  
  # Get reachcodes for waterbody outlets, so we have an idea of which
  #     NHD HR 4-digit geodatabase we may need to retrieve
  RC <- filter(nhd, COMID %in% nhd_hr_wb_resops$resops_flowlcomid)$REACHCODE
  
  # If no NHDArea or HR waterbodies needed return NULL
  if (nrow(nhd_area_resops) == 0 & nrow(nhd_hr_wb_resops) == 0){
    return(NA)
  }
  
  # If NHDArea feature needed retrieve from National GDB
  if (nrow(nhd_area_resops) > 0){
    nhd_area_vpu <- nhdplus_area %>%
      filter(COMID %in% nhd_area_resops$member_comid) %>%
      mutate(source = "NHDv2Area")
    
    wb <- st_transform(nhd_area_vpu, crs)
  }
  
  # If NHDHR feature needed
  if (nrow(nhd_hr_wb_resops) > 0){
    # HUC04 we need to download
    huc04 <- substr(RC, 1, 4)
    
    # Download NHD HR HUC04 if we dont' have it, other wise load and
    # Bind NHDHR waterbodies into one layer
    hr_wb <- do.call("rbind", lapply(unique(huc04), function(x){
      print(x)
      vpu <- unique(substr(x, 1, 2))
      nhd_file <- list.files(nhdplushr_dir, pattern = paste0(".*",x,".*\\.gdb$"), 
                             recursive = TRUE, full.names = TRUE,include.dirs = TRUE)
      
      if(length(nhd_file) < 1) {
        download_nhdplushr(nhdplushr_dir, unique(huc04))
      }
      
      nhd_file <- list.files(nhdplushr_dir, pattern = paste0(".*",x,".*\\.gdb$"), 
                             recursive = TRUE, full.names = TRUE,include.dirs = TRUE)
      
      if(length(nhd_file) != 1) {
        stop("something wrong with nhdplushr file in ", x)
      }
      
      # Format to similar to NHDArea/Waterbody layers
      read_sf(nhd_file, "NHDWaterbody")})) #%>%
    
    # convert to lower case and designate new shape field
    names(hr_wb) <- tolower(names(hr_wb)) 
    st_geometry(hr_wb) <- "shape"
    
    hr_wb <- filter(hr_wb, permanent_identifier %in% WBs$member_comid) %>%
      rename(GNIS_NAME = gnis_name, GNIS_ID = gnis_id, 
             Permanent_Identifier = permanent_identifier, AREASQKM = areasqkm, 
             FTYPE = ftype, FCODE = fcode, FDATE = fdate, 
             REACHCODE = reachcode) %>%
      #select(-c(permanent_identifier, visibilityfilter, vpuid)) %>%
      st_zm(.) %>%
      st_as_sf() %>%
      mutate(source = "NHDHR",
             wb_id = ifelse(!is.na(GNIS_ID), GNIS_ID, Permanent_Identifier)) %>%
      st_transform(crs)
    
    # Bind or create new object
    if(exists("wb")){
      hr_wb <- st_compatibalize(hr_wb, wb)
      
      wb <- data.table::rbindlist(list(wb, hr_wb), fill = TRUE) %>%
        st_as_sf()
    } else {
      wb <- hr_wb
    }
    
  }
  
  # # get the outlt rows from the table
  # resops_outlet <- read.csv("data/reservoir_data/ISTARF-CONUS.csv") %>%
  #    filter(WBAREACOMI %in% wb$COMID | HR_NHDPLUSID %in% wb$COMID)
  
  # Clear out columns not needed
  WBs_sub <- WBs %>%
    select(-any_of(c("COMID", "GNIS_ID")))
  
  WBs_fin <- st_drop_geometry(WBs_sub) %>%
    mutate(member_comid = member_comid) %>%
    inner_join(select(wb, Permanent_Identifier, GNIS_ID, GNIS_NAME2 = GNIS_NAME), 
               by = c("member_comid" = "Permanent_Identifier")) %>%
    mutate(wb_id = ifelse(GNIS_ID %in% c(NA, " "), member_comid, GNIS_ID), 
           GNIS_NAME = ifelse(is.na(GNIS_NAME), GNIS_NAME2, GNIS_NAME)) %>%
    select(-c(GNIS_ID, GNIS_NAME2)) %>%
    st_as_sf() 
  
  # Create table of all flowlines that intersect the waterbody
  nhd_wb <- st_intersects(st_transform(nhd, st_crs(wb)), wb) 
  comid <- nhd[lengths(nhd_wb) > 0,]$COMID
  nhd_wb_all <- nhd_wb[lengths(nhd_wb) > 0] %>%
    purrr::set_names(comid) %>%
    stack() %>%
    # Ind is the default name for the set_names
    rename(comid = ind, nhd_wb = values) %>%
    mutate(wb_comid = wb[nhd_wb,]$Permanent_Identifier,
           outlet = ifelse(comid %in% WBs_fin$resops_flowlcomid, "outlet", NA),
           comid = as.integer(as.character(comid))) %>%
    #left_join(select(wb_table, DAM_ID, DAM_NAME, resops_FlowLcomid), 
    #          by = c("comid" = "resops_FlowLcomid")) %>%
    left_join(select(st_drop_geometry(wb), wb_id, Permanent_Identifier, GNIS_ID, 
                     source), 
              by = c("wb_comid" = "Permanent_Identifier"))
  
  return(list(nhd_wb_table = nhd_wb_all, wb = WBs_fin))
}

### USEDIN wbout_POI_creaton and wbin_POIcreation
#'  Creates wb inlet and outlet events for splitting in hyRefactor
#'          for waterbodies derived from NHDArea and NHDHR waterbodies
#'  @param WBs  (sf data.frame) return from HR_Area_coms
#'  @param nhd_wb (sf data.frame) flowlines that intersect waterbodies
#'  @param type (character) whether to derive inlet or outlet points
#' 
#'  @return (sf data.frame) dataframe of WB inlet and outlet points to split
#'  
WB_event <- function(WBs, nhd_wb, type){
  # split into features and table
  WBs_table <- WBs$nhd_wb_table
  WBs_layer <- WBs$wb
  
  if (type == "outlet"){
    # get the outlet comid from the ResOps Table
    outlet_fl <- filter(nhd_wb, COMID %in% filter(WBs_table, outlet == "outlet")$comid)
    
    # Get the downstream flowline for continuity
    ds_fl <- filter(nhd_wb, Hydroseq %in% outlet_fl$DnHydroseq,
                    LevelPathI %in% outlet_fl$LevelPathI) %>%
      rbind(outlet_fl) %>%
      arrange(desc(Hydroseq)) %>%
      group_by(LevelPathI) %>%
      # union together 
      summarize(do_union = TRUE) %>%
      st_cast("LINESTRING")
    
    WBs_HR <- filter(WBs_layer, source == "HR ID AVAILABLE")
    
    # For NHD HR waterbody outlets its a bit more complicated
    if (nrow(WBs_HR) > 0){
      # Get the flowlines intersecting the HR waterbody and find one with the
      #     max DA
      outlet_wb_int <- nhd_wb[lengths(st_intersects(st_transform(nhd_wb, st_crs(WBs_HR)),
                                                    WBs_HR)) > 0,] %>%
        group_by(wb_id) %>%
        filter(TotDASqKM == max(TotDASqKM)) %>%
        ungroup()
      
      # get the ds flo with the same levepath (JIC)
      ds_fl <- filter(nhd_wb, DnHydroseq %in% outlet_wb_int$Hydroseq,
                      LevelPathI %in% outlet_wb_int$LevelPathI)
      
      outlet_fl <- rbind(outlet_wb_int, ds_fl)
      
      # Cast flowlines within NHDHR waterbody to point
      WB_FL_pnts <- outlet_wb_int %>%
        st_cast("POINT") %>%
        group_by(wb_id) %>%
        mutate(pnt_id = row_number()) %>%
        ungroup()
      
      outlet_pnts <- do.call("rbind", lapply(unique(WB_FL_pnts$wb_id), function(x){
        fl <- filter(WB_FL_pnts, wb_id == x)
        wb <- filter(WBs_HR, wb_id == x)
        
        # Determine which points intersect waterbody
        WB_outlet_pnts <- fl[lengths(st_intersects(fl, wb)) > 0,] %>%
          st_drop_geometry() %>%
          group_by(wb_id) %>%
          mutate(within_wb_id = row_number()) %>%
          filter(within_wb_id >= max(within_wb_id)) %>%
          ungroup() %>%
          select(wb_id, orig_pnt_id = pnt_id, within_wb_id)
        
        # Deriv new linestring by concating points from most upstream point
        #       within waterbody to downstream so we can split at FL/waterbody
        #       nexus
        outlet_FL <- fl %>%
          inner_join(WB_outlet_pnts, by = "wb_id") %>%
          select(wb_id, pnt_id, orig_pnt_id, within_wb_id) %>%
          filter(pnt_id >= orig_pnt_id) %>%
          group_by(wb_id) %>%
          summarize(do_union = F) %>%
          st_cast("LINESTRING") %>%
          filter(wb_id %in% wb$wb_id)
        
        outlet_pnt <- sf::st_intersection(outlet_FL, WBs_HR) %>%
          st_cast("MULTIPOINT") %>%
          st_cast("POINT") %>% # was point
          group_by(wb_id) %>%
          filter(row_number() == max(row_number(), na.rm = T)) %>%
          ungroup() %>%
          mutate(id = row_number()) %>%
          filter(wb_id == wb_id.1)
      }))
      
      # Derive the events
      if(exists("wb_events")){
        hr_events <- get_flowline_index(st_transform(nhd_wb, st_crs(outlet_pnts)),
                                        outlet_pnts) %>%
          inner_join(select(st_drop_geometry(nhd_wb), COMID, wb_id), by = "COMID") %>%
          filter(wb_id %in% WBs_HR$wb_id) %>%
          mutate(event_type = type) %>%
          cbind(select(outlet_pnts, geom)) %>%
          st_as_sf() %>%
          st_transform(st_crs(wb_events)) %>%
          st_compatibalize(., wb_events)
        
        wb_events <- rbind(wb_events, hr_events) %>%
          distinct()
      } else {
        wb_events <- do.call("rbind", lapply(unique(outlet_pnts$wb_id), function(x){
          outlet_pnt <- filter(outlet_pnts, wb_id == x)
          get_flowline_index(nhd_wb, outlet_pnt)})) %>%
          left_join(distinct(st_drop_geometry(outlet_pnts), resops_flowlcomid, wb_id), 
                    by = c("COMID" = "resops_flowlcomid")) %>%
          mutate(event_type = type) %>%
          cbind(select(outlet_pnts, geom)) %>%
          st_as_sf()
      }
      
    }
    
    # For inlet points its alot easier for both NHDARea and NHDHR
  } else {
    start_pts <- get_node(nhd_wb, position = "start") %>%
      cbind(st_drop_geometry(nhd_wb))
    
    inlet_FL <- nhd_wb[lengths(st_intersects(
      st_transform(start_pts, st_crs(WBs_layer)), WBs_layer)) == 0,] %>%
      rbind(filter(nhd_wb, Hydroseq %in% .$DnHydroseq, 
                   LevelPathI %in% .$LevelPathI)) %>%
      arrange(desc(Hydroseq)) %>%
      unique()
    
    inlet_ls <- inlet_FL %>%
      group_by(LevelPathI) %>%
      st_cast("POINT") %>%
      summarize(do_union = F) %>%
      st_cast("LINESTRING") %>%
      ungroup()
    
    inlet_pnts_int <- sf::st_intersection(
      st_transform(inlet_ls, st_crs(WBs_layer)), WBs_layer) #%>%
    #st_cast("LINESTRING")
    
    int_point <- inlet_pnts_int[st_geometry_type(inlet_pnts_int) %in% c("POINT"),]
    if(nrow(int_point) > 0){
      inlet_fl <- inlet_pnts_int[st_geometry_type(inlet_pnts_int) %in% c("LINESTRING", "MULTILINESTRING"),] %>%
        st_cast("LINESTRING")
      inlet_pnts_int <- rbind(int_point, inlet_fl)
    }
    
    inlet_pnts <- inlet_pnts_int %>%
      group_by(LevelPathI) %>%
      st_cast("POINT") %>%
      mutate(id = row_number()) %>%
      filter(id == min(id), !is.na(wb_id)) %>%
      ungroup() 
    
    wb_events <- get_flowline_index(st_transform(nhd_wb, st_crs(inlet_pnts)),
                                    inlet_pnts) %>%
      inner_join(select(st_drop_geometry(nhd_wb), COMID, wb_id, LevelPathI), by = "COMID") %>%
      mutate(event_type = type) %>%
      inner_join(select(inlet_pnts, LevelPathI), by = "LevelPathI") %>%
      #group_by(COMID, LEVELPATHI) %>%
      mutate(nexus = T) %>%
      rename(POI_identifier = wb_id) %>%
      select(-c(id, offset)) %>%
      st_as_sf()
  }
  return(wb_events)
} 

### used as part of waterbody poi vreation in the workflow
#'  Collapses POIs us/ds of waterbody POIs
#'  @param POIs (sf data.frame) rolling POI data frame
#'  @param nhd (sf data.frame) nhd flowlines 
#'  @param events (sf data.frame) waterbody events
#'  @param gages gages from gage selection
#' 
#'  @return (sf data.frame) dataframe of wb inlet POIs collapsed
wb_poi_collapse <- function(POIs, nhd, events, gages){
  gage_dist_node <- function(x, wb_ds_ds, gage_add, events){
    print (x) 
    wb_out_fl <- distinct(filter(wb_ds_ds, COMID == x))
    gage_ds <- filter(wb_ds_ds, Hydroseq %in% wb_out_fl$Hydroseq |
                        Hydroseq %in% wb_out_fl$DnHydroseq) 
    
    gage_reach <- gage_ds %>%
      group_by(REACHCODE) %>%
      summarize(do_union = TRUE,
                total_length = sum(LENGTHKM))
    
    #print(nrow(gage_reach))
    gage_event <- filter(gage_add, COMID %in% gage_ds$COMID) %>%
      filter(reach_meas == min(reach_meas))
    wb_event <- filter(events, COMID == x, event_type == "outlet") %>%
      unique()
    
    if(nrow(gage_reach) == 0){
      print("no gages")
    }
    
    if(nrow(gage_event) == 0){
      return("no events")
    } else if(gage_event$COMID != wb_out_fl$COMID) {
      gage_reach <- gage_reach %>%
        filter(REACHCODE == unique(gage_event$reachcode)) %>%
        mutate(gage_dist = ifelse(gage_event$nexus == TRUE,
                                  total_length * (1 - (gage_event$reach_meas/100)),
                                  total_length)) %>%
        mutate(gage_comid = gage_event$COMID,
               wbout_comid = x)
    } else if(gage_event$COMID == wb_out_fl$COMID){
      if(nrow(wb_event) >0){
        wb_out_meas <- min(wb_event$REACH_meas)
        wb_RC <- wb_event$REACHCODE
      } else {
        wb_out_meas <- min(wb_out_fl$FromMeas)
        wb_RC <- wb_out_fl$REACHCODE
      }
      
      # gage reach
      gage_reach <- gage_reach %>%
        filter(REACHCODE == gage_event$reachcode) %>%
        mutate(gage_dist = total_length * (1 - (gage_event$reach_meas/100))) 
      
      # wb info
      wb_reach <- gage_reach %>%
        filter(REACHCODE == unique(wb_RC)) %>%
        mutate(wb_dist = total_length * (1 - (wb_out_meas /100)))
      
      gage_reach <- gage_reach %>%
        mutate(gage_dist = abs(wb_reach$wb_dist - gage_dist),
               gage_comid = gage_event$COMID,
               wbout_comid = x)
    }
  }
  
  # Previously identified streamgages within Gage_Selection.Rmd
  streamgages_VPU <- gages %>%
    rename(COMID = comid) %>%
    filter(COMID %in% nhd$COMID) %>%
    switchDiv(., nhd) 
  
  # get waterbody outlets
  wb_out <- filter(POIs, !is.na(Type_WBOut), is.na(Type_Gages))
  
  wb_out_nexus <- filter(wb_out, nexus)
  wb_out_nhd <- filter(nhd, COMID %in% wb_out$COMID) 
  wb_ds_ds <- wb_out_nhd %>%
    rbind(filter(nhd, LevelPathI %in% .$LevelPathI & Hydroseq %in% .$DnHydroseq))
  
  # get gages
  gage_wb <- filter(POIs, !is.na(Type_Gages) & is.na(Type_WBOut)) #%>%
  #  filter(COMID %in% wb_ds_ds$COMID) 
  #gage_nhd <- filter(nhd, COMID %in% gage_wb$COMID)
  #gage_node <- filter(gage_wb, nexus == FALSE)
  #gage_nexus <- filter(gage_wb, nexus == TRUE)
  
  gage_add <- filter(streamgages_VPU, site_no %in% gage_wb$Type_Gages) %>%
    select(COMID, reachcode, reach_meas, site_no) %>%
    inner_join(select(st_drop_geometry(gage_wb), site_no = Type_Gages, nexus), 
               by = "site_no") %>%
    distinct()
  
  # 8693865
  output <- lapply(wb_out$COMID, 
                   gage_dist_node, wb_ds_ds, gage_add, events)
  output_length <- output[lengths(output) > 1]
  
  if(length(output_length) == 0){
    return(list(POIs = POIs, events_ret = NA))
  }
  
  output_full <- do.call("rbind", output[lengths(output) > 1]) %>%
    filter(gage_dist < 1) %>%
    left_join(select(st_drop_geometry(nhd), COMID, TotDASqKM_WB = TotDASqKM),
              by = c("wbout_comid" = "COMID")) %>%
    left_join(select(st_drop_geometry(nhd), COMID, TotDASqKM_gage = TotDASqKM),
              by = c("gage_comid" = "COMID")) %>%
    mutate(DAR = TotDASqKM_WB / TotDASqKM_gage) %>%
    filter(gage_dist < 1, DAR > .975) %>%
    st_drop_geometry()
  
  gage_POI <- filter(POIs, COMID %in% output_full$gage_comid) %>%
    select(COMID, Type_HUC12_ds = Type_HUC12, Type_Gages_ds = Type_Gages, 
           Type_TE_ds = Type_TE, nexus) %>%
    st_drop_geometry() %>%
    group_by(COMID) %>%
    summarise(Type_HUC12_ds = last(na.omit(Type_HUC12_ds)), 
              Type_Gages_ds = last(na.omit(Type_Gages_ds)),
              Type_TE_ds = last(na.omit(Type_TE_ds)),
              nexus = last(na.omit(nexus)))
  
  WB_POI <- filter(POIs, COMID %in% output_full$wbout_comid, !is.na(Type_WBOut)) %>%
    inner_join(select(output_full, gage_comid, wbout_comid), by = c("COMID" = "wbout_comid")) %>%
    inner_join(select(gage_POI, -nexus), by = c("gage_comid" = "COMID")) %>%
    mutate(Type_HUC12 = ifelse(!is.na(Type_HUC12_ds), Type_HUC12_ds, Type_HUC12),
           Type_Gages = ifelse(!is.na(Type_Gages_ds), Type_Gages_ds, Type_Gages),
           Type_TE = ifelse(!is.na(Type_TE_ds), Type_TE_ds, Type_TE)) %>%
    select(-c(Type_HUC12_ds, Type_Gages_ds, Type_TE_ds))
  
  POIs_fin <- filter(POIs, !COMID %in% c(WB_POI$COMID, WB_POI$gage_comid)) %>%
    rbind(select(WB_POI, -gage_comid))
  
  if(any(gage_POI$nexus == TRUE)){
    gage_events <- filter(gage_POI, nexus == TRUE)
    events <- filter(events, !POI_identifier %in% gage_events$Type_Gages_ds)
  }
  
  
  return(list(POIs = POIs_fin, events_ret = events))
}

#' create abstraction(?) and removal event points of interest
#' @param ar_data abstraction and removal data from USGS data release
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
create_ar_event_pois <- function(POIs, ar_data, flowline, poi_name = "AR") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  ar_events <- ar_data %>%
    filter(fromComid > 0 | ToComid > 0) %>%
    select(COMID = NHDComid, AREventID) %>%
    filter(COMID %in% flowline$COMID)
  
  if(nrow(ar_events) > 0){
    # AR POIs
    POIs <- POI_creation(ar_events, flowline, poi_name) %>%
      addType(., POIs, "AR", nexus = TRUE)
    
  }

  list(POIs = POIs, vpu = vpu)  
}

#' create terminal pois
#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
#' @param min_da_km_terminal minimum drainage area to include a network
#' @param poi_name name for poi set
create_terminal_pois <- function(POIs, flowline, min_da_km_terminal, poi_name = "Term") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  term_paths <- filter(st_drop_geometry(flowline), 
                       Hydroseq %in% filter(flowline, COMID %in% POIs$COMID)$TerminalPa)
  
  # Non-POI levelpath terminal pois, but meet size criteria
  terminal_POIs <- st_drop_geometry(flowline) %>%
    filter(Hydroseq == TerminalPa | (toCOMID == 0 | is.na(toCOMID) | !toCOMID %in% COMID)) %>%
    filter(!COMID %in% term_paths$COMID, TotDASqKM >= min_da_km_terminal) %>%
    bind_rows(term_paths) %>%
    # Use level path identifier as Type ID
    select(COMID, LevelPathI)
  
  POIs <- POI_creation(terminal_POIs, flowline, poi_name) %>%
    addType(., POIs, poi_name, nexus = TRUE)
 
  list(POIs = POIs, vpu = vpu) 
}

#' create confluence pois
#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
create_confluence_pois <- function(POIs, flowline, poi_name = "Conf") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  # Navigate upstream from each POI and determine minimally-sufficient network between current POI sets
  up_net <- unique(unlist(lapply(unique(POIs$COMID), 
                                 \(x, nhdDF) get_UM(nhdDF, x), 
                                 nhdDF = flowline)))
  
  finalNet <- unique(NetworkConnection(up_net, st_drop_geometry(flowline)))
  
  # Subset NHDPlusV2 flowlines to navigation results and write to shapefile
  flowline <- mutate(flowline, minNet = ifelse(COMID %in% finalNet, 1, 0)) 
  
  # Create new confluence POIs
  conf_COMIDs <- st_drop_geometry(filter(flowline, minNet == 1)) %>%
    # Downstream hydrosequence of 0 indicates
    #   the flowline is terminating or
    #   leaving the domain, so they
    #   are excluded from this process
    filter(DnHydroseq > 0) %>%
    group_by(DnHydroseq) %>%
    filter(n()> 1) %>%
    mutate(Type_Conf = LevelPathI) %>%
    ungroup() %>%
    select(COMID, Type_Conf)
  
  POIs <- POI_creation(conf_COMIDs, filter(flowline, minNet == 1), poi_name) %>%
    addType(., POIs, poi_name, nexus = TRUE)
  
  list(POIs = POIs, updated_flowline = flowline, vpu = vpu)
}

#' create waterbody inlet pois
#' @param POIs POIs from previous step
#' @param wb_lyr waterbodies to create inlet pois from
#' @param events complete layer of events prior to creating waterbody inlets
#' @param flowline flowlines to attach pois to
#' @param nhdplus_area (sf data.frame) NHDPlus Area features
#' @param proj_crs projected crs for overall project
#' @param poi_name name for poi set
#' @param nhdplushr_dir directory containing NHDPlusHR data
create_wb_inlet_pois <- function(POIs, wb_lyr, events, 
                                 flowline, gages, nhdplus_area, 
                                 proj_crs, poi_name = "WBIn", nhdplushr_dir) {

  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  wb_layers <- wbin_POIcreation(filter(flowline, minNet == 1), wb_lyr, events,
                                nhdplus_area, proj_crs, POIs, 
                                poi_name, nhdplushr_dir)
  
  wb_in_col <- wb_inlet_collapse(wb_layers$POIs, flowline, events, gages)
  
  wb_inlet_events <- NULL
  
  if(!all(is.na(wb_layers$events))) {
    wb_inlet_events <- wb_layers$events %>%
      select(COMID, REACHCODE, REACH_meas, POI_identifier,
             event_type, nexus)
    
    events <- rbind(events, st_compatibalize(wb_inlet_events, events))
  }
  
  list(POIs = wb_in_col$POIs, events = events, vpu = vpu)
}

### used as part of waterbody poi creation in the workflow
#'  collapses close upstream gages with wb inlet events.
#'  @param POIs (sf data.frame) rolling POI data frame
#'  @param nhd (sf data.frame) nhd flowlines
#'  @param events (list) waterbody inlet events
#' 
#'  @return (sf data.frame) dataframe of gage POIs
#' 
wb_inlet_collapse <- function(POIs, nhd, events, gages){
  gage_dist_node <- function(x, wb_ds_ds, gage_add, events){
    # print (x)
    wb_in_fl <- filter(wb_ds_ds, COMID == x)
    gage_ds <- filter(wb_ds_ds, Hydroseq %in% wb_in_fl$Hydroseq |
                        Hydroseq %in% wb_in_fl$DnHydroseq) %>%
      filter(TotDASqKM == max(TotDASqKM))
    
    gage_reach <- gage_ds %>%
      group_by(REACHCODE) %>%
      summarize(do_union = TRUE,
                total_length = sum(LENGTHKM))
    
    #print(nrow(gage_reach))
    gage_event <- filter(gage_add, COMID %in% gage_ds$COMID)
    wb_event <- filter(events, COMID == x, event_type == "inlet") %>%
      unique()
    
    if(nrow(gage_reach) == 0){
      print("no gage reaches")
    }
    
    if(nrow(gage_event) == 0){
      return("No gage events")
    } else if(gage_event$COMID != wb_in_fl$COMID) {
      gage_reach <- gage_reach %>%
        filter(REACHCODE == gage_event$reachcode) %>%
        mutate(gage_dist = ifelse(gage_event$nexus == TRUE,
                                  total_length * (1 - (gage_event$reach_meas/100)),
                                  total_length)) %>%
        mutate(gage_comid = gage_event$COMID,
               wbin_comid = x)
    } else if(gage_event$COMID == wb_in_fl$COMID){
      if(nrow(wb_event) >0){
        wb_in_meas <- wb_event$REACH_meas
        wb_RC <- wb_event$REACHCODE
      } else {
        wb_out_meas <- wb_in_fl$FromMeas
        wb_RC <- wb_in_fl$REACHCODE
      }
      
      # gage reach
      gage_reach <- gage_reach %>%
        filter(REACHCODE == gage_event$reachcode) %>%
        mutate(gage_dist = total_length * (1 - (gage_event$reach_meas/100))) 
      
      # wb info
      wb_reach <- gage_reach %>%
        filter(REACHCODE == wb_RC) %>%
        mutate(wb_dist = total_length * (1 - (wb_out_meas /100)))
      
      gage_reach <- gage_reach %>%
        mutate(gage_dist = abs(wb_reach$wb_dist - gage_dist),
               gage_comid = gage_event$COMID,
               wbin_comid = x)
    }
  }
  
  #events <- read_sf(temp_gpkg, split_layer) %>%
  #  rbind(st_compatibalize(wb_,.))
  
  # Previously identified streamgages within Gage_Selection.Rmd
  streamgages_VPU <- gages %>%
    rename(COMID = comid) %>%
    filter(COMID %in% nhd$COMID) %>%
    #st_drop_geometry() %>%
    switchDiv(., nhd) 
  
  # get waterbody outlets
  wb_in <- filter(POIs, !is.na(Type_WBIn), is.na(Type_Gages))
  
  wb_in_nexus <- filter(wb_in, nexus == TRUE)
  wb_in_nhd <- filter(nhd, COMID %in% wb_in$COMID) 
  wb_ds_ds <- wb_in_nhd %>%
    rbind(filter(nhd, LevelPathI %in% .$LevelPathI & DnHydroseq %in% .$Hydroseq)) %>%
    distinct()
  
  # get gages
  gage_wb <- filter(POIs, !is.na(Type_Gages) & is.na(Type_WBIn)) #%>%
  #  filter(COMID %in% wb_ds_ds$COMID) 
  #gage_nhd <- filter(nhd, COMID %in% gage_wb$COMID)
  #gage_node <- filter(gage_wb, nexus == FALSE)
  #gage_nexus <- filter(gage_wb, nexus == TRUE)
  
  gage_add <- filter(streamgages_VPU, site_no %in% gage_wb$Type_Gages) %>%
    select(COMID, reachcode, reach_meas, site_no) %>%
    inner_join(select(st_drop_geometry(gage_wb), site_no = Type_Gages, nexus), 
               by = "site_no") %>%
    filter(!nexus)
  
  output <- lapply(wb_in$COMID, 
                   gage_dist_node, wb_ds_ds, gage_add, events)
  
  output_full <- do.call("rbind", output[lengths(output) > 1]) 
  
  if(!is.null(output_full)){
    output_full <- output_full %>%
      filter(gage_dist < 1) %>%
      left_join(select(st_drop_geometry(nhd), COMID, TotDASqKM_WB = TotDASqKM),
                by = c("wbin_comid" = "COMID")) %>%
      left_join(select(st_drop_geometry(nhd), COMID, TotDASqKM_gage = TotDASqKM),
                by = c("gage_comid" = "COMID")) %>%
      mutate(DAR = TotDASqKM_gage / TotDASqKM_WB) %>%
      filter(gage_dist < 1, DAR > .975) %>%
      st_drop_geometry()
    
    gage_POI <- filter(POIs, COMID %in% output_full$gage_comid) %>%
      select(COMID, Type_HUC12_ds = Type_HUC12, Type_Gages_ds = Type_Gages, 
             Type_TE_ds = Type_TE, Type_Term_ds = Type_Term, nexus) %>%
      st_drop_geometry() %>%
      group_by(COMID) %>%
      summarise(Type_HUC12_ds = last(na.omit(Type_HUC12_ds)), 
                Type_Gages_ds = last(na.omit(Type_Gages_ds)),
                Type_Term_ds = last(na.omit(Type_Term_ds)),
                Type_TE_ds = last(na.omit(Type_TE_ds)),
                nexus = last(na.omit(nexus)))
    
    WB_POI <- filter(POIs, COMID %in% output_full$wbin_comid, !is.na(Type_WBIn)) %>%
      left_join(select(output_full, gage_comid, wbin_comid), by = c("COMID" = "wbin_comid")) %>%
      left_join(select(gage_POI, -nexus), by = c("gage_comid" = "COMID")) %>%
      mutate(Type_HUC12 = ifelse(!is.na(Type_HUC12_ds), Type_HUC12_ds, Type_HUC12),
             Type_Gages = ifelse(!is.na(Type_Gages_ds), Type_Gages_ds, Type_Gages),
             Type_TE = ifelse(!is.na(Type_TE_ds), Type_TE_ds, Type_TE),
             Type_Term = ifelse(!is.na(Type_Term_ds), Type_Term_ds, Type_Term)) %>%
      select(-c(Type_HUC12_ds, Type_Gages_ds, Type_TE_ds, Type_Term_ds))
    
    POIs_fin <- filter(POIs, !COMID %in% c(WB_POI$COMID, WB_POI$gage_comid)) %>%
      rbind(select(WB_POI, -gage_comid))
    
    if(any(gage_POI$nexus == TRUE)){
      gage_events <- filter(gage_POI, nexus == TRUE)
      events <- filter(events, !POI_identifier %in% gage_events$Type_Gages_ds)
    }
    return(list(POIs = POIs_fin, events_ret = events))
  } else {
    print ("no points collapse")
    return(list(POIs = POIs, events_ret = NA))
  }
}

#' @param POIs POIs from previous step
#' @param nid_points national inventory of dams data
#' @param flowline flowlines to attach pois to
#' @param poi_name name for poi set
create_nid_pois <- function(POIs, nid_points, flowline, poi_name) {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  # Read in NID shapefile
  NID_COMIDs <- nid_points %>%
    st_drop_geometry() %>%
    filter(EROM != 0, FlowLcomid %in% filter(flowline, dend ==1)$COMID) %>%
    rename(COMID = FlowLcomid) %>%
    switchDiv(., flowline) %>%
    group_by(COMID) %>%
    summarize(Type_NID = paste0(unique(NIDID), collapse = " "))
  
  # Derive other NID POIs
  POIs <- POI_creation(NID_COMIDs, flowline, poi_name) %>%
    addType(., POIs, poi_name, nexus = FALSE, bind = FALSE)
  
  list(POIs = POIs, vpu = vpu)
}

#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
#' @param all_nhdplus_attributes original attributes from nhdplusv2
#' @param min_da_km_hw minimum drainage area for creation of a headwater poi
#' @param poi_name name for poi set
create_headwater_pois <- function(POIs, flowline, 
                                  all_nhdplus_attributes, 
                                  min_da_km_hw, poi_name = "DA") {

  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  # derive incremental segments from POIs
  inc_segs <- make_incremental_segments(flowline,
                                        filter(st_drop_geometry(flowline),
                                               COMID %in% POIs$COMID)) %>%
    # bring over VAA data
    inner_join(select(all_nhdplus_attributes, COMID, 
                      DnHydroseq, VA_MA, TOTMA, LENGTHKM, MAXELEVSMO,
                      MINELEVSMO, WBAREACOMI, WBAreaType, FTYPE, StartFlag,
                      AreaSqKM, TotDASqKM), by = "COMID")
  
  hw_segs <- inc_segs %>%
    group_by(POI_ID) %>%
    filter(any(StartFlag == 1)) %>%
    filter(any(TotDASqKM > max_da_km_hw)) %>%
    ungroup()
  
  att_group <- function(a, athres) {
    #cumsum <- 0
    group  <- 1
    result <- numeric()
    for (i in 1:length(a)) {
      #cumsum <- cumsum + a[i]
      tot_DA <- a[i]
      if (tot_DA > athres) {
        group <- group + 1
        athres <- athres + athres
      }
      result = c(result, group)
    }
    return (result)
  }  
  
  #TODO Add magic numbers to config file
  hw_DA_splits <- hw_segs %>%
    st_drop_geometry() %>%
    #filter(.data$ID %in% cat$ID) %>%
    group_by(POI_ID) %>%
    arrange(-Hydroseq) %>%
    mutate(ind = att_group(TotDASqKM, min_da_km_hw)) %>%
    ungroup() %>%
    group_by(POI_ID, ind) %>%
    mutate(total_length = cumsum(LENGTHKM)) %>%
    ungroup() %>%
    group_by(POI_ID, ind) %>%
    mutate(set = cur_group_id()) %>%
    filter(TotDASqKM == max(TotDASqKM) & 
             total_length > 5) %>%
    mutate(da_class = paste(POI_ID, ind, sep = "_")) %>%
    ungroup() %>%
    select(-ind) %>%
    filter(!COMID %in% POIs$COMID)
  
  if(nrow(hw_DA_splits) > 0) {
    POIs <- POI_creation(select(hw_DA_splits, COMID, da_class), 
                             filter(flowline, poi == 1), poi_name) %>%
      addType(., POIs, poi_name, nexus = TRUE)
  }
  
  list(POIs = POIs, vpu = vpu)
  
}

#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
#' @param all_nhdplus_attributes original attributes from nhdplusv2
#' @param poi_name name for poi set
create_elevation_break_pois <- function(POIs, flowline, all_nhdplus_attributes,
                                        poi_name = "elev") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs
  
  inc_segs <- make_incremental_segments(filter(flowline, minNet == 1), 
                                        filter(st_drop_geometry(flowline),
                                               COMID %in% POIs$COMID)) %>%
    # bring over VAA data
    inner_join(select(all_nhdplus_attributes, COMID, 
                      DnHydroseq, VA_MA, TOTMA, LENGTHKM, MAXELEVSMO,
                      MINELEVSMO, WBAREACOMI, WBAreaType, FTYPE, StartFlag,
                      AreaSqKM, TotDASqKM), by = "COMID")
  
  elev_fp <- inc_segs %>%
    group_by(POI_ID) %>%
    arrange(Hydroseq) %>%
    # Get elevation info
    mutate(MAXELEVSMO = na_if(MAXELEVSMO, -9998), MINELEVSMO = na_if(MINELEVSMO, -9998), 
           elev_diff_seg = max(MAXELEVSMO) - min(MINELEVSMO),
           total_length = cumsum(LENGTHKM)) %>%
    filter((max(MINELEVSMO) - min(MINELEVSMO)) > elev_diff) %>%
    mutate(inc_elev_diff = c(MINELEVSMO[1], (MINELEVSMO - lag(MINELEVSMO))[-1])) %>%
    mutate(inc_elev_diff = ifelse(inc_elev_diff == MINELEVSMO, 0, inc_elev_diff)) %>%
    ungroup()
  
  if(nrow(elev_fp) > 0){
    #TODO Add magic numbers to config file
    hw_elev_splits <- elev_fp %>%
      st_drop_geometry() %>%
      group_by(POI_ID) %>%
      arrange(-Hydroseq) %>%
      mutate(ind = cs_group(inc_elev_diff, elev_diff/2)) %>%
      ungroup() %>%
      group_by(POI_ID, ind) %>%
      filter(TotDASqKM == max(TotDASqKM) &
               total_length > 5) %>%
      mutate(elev_class = paste(POI_ID, ind, sep = "_")) %>%
      ungroup() %>%
      select(-ind) %>%
      filter(!COMID %in% POIs$COMID)
    
    if(nrow(hw_elev_splits) > 0) {
      POIs <- POI_creation(select(hw_elev_splits, COMID, elev_class), 
                               filter(flowline, poi == 1),
                               poi_name) %>%
        addType(., POIs, poi_name, nexus = TRUE)
    }
  }
  
  list(POIs = POIs, vpu = vpu)
}

cs_group <- function(a, athres) {
  cumsum <- 0
  group  <- 1
  result <- numeric()
  for (i in 1:length(a)) {
    cumsum <- cumsum + a[i]
    if (cumsum > athres) {
      group <- group + 1
      cumsum <- a[i]
    }
    result = c(result, group)
  }
  return (result)
}  

#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
#' @param all_nhdplus_attributes original attributes from nhdplusv2
#' @param poi_name name for poi set
create_time_of_travel_pois <- function(POIs, flowline, all_nhdplus_attributes, poi_name = "Travel") {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs

  # derive incremental segments from POIs
  inc_segs <- make_incremental_segments(flowline,
                                        filter(st_drop_geometry(flowline),
                                               COMID %in% POIs$COMID, 
                                               COMID %in% flowline$COMID)) %>%
    # bring over VAA data
    inner_join(select(all_nhdplus_attributes, COMID, DnHydroseq, VA_MA, TOTMA, LENGTHKM,
                      MAXELEVSMO, MINELEVSMO, WBAREACOMI, WBAreaType, FTYPE,
                      AreaSqKM, TotDASqKM), by = "COMID")
  
  # TT POIs
  tt_pois_split <- inc_segs %>%
    arrange(-Hydroseq) %>%
    # Should we substitute with a very small value
    mutate(VA_MA = ifelse(VA_MA < 0, NA, VA_MA)) %>%
    mutate(FL_tt_hrs = (LENGTHKM * ft_per_km)/ VA_MA / s_per_hr ) %>%
    group_by(POI_ID) %>%
    filter(sum(FL_tt_hrs) > travt_diff, max(TotDASqKM) > max_elev_TT_DA) %>%
    mutate(cum_tt = cumsum(FL_tt_hrs),
           hrs = sum(FL_tt_hrs)) 
  
  #TODO Add magic numbers to config file
  tt_splits <- tt_pois_split %>%
    st_drop_geometry() %>%
    group_by(POI_ID) %>%
    arrange(-Hydroseq) %>%
    mutate(ind = cs_group(FL_tt_hrs, travt_diff * 0.75)) %>%
    ungroup() %>%
    group_by(POI_ID, ind) %>%
    filter(TotDASqKM == max(TotDASqKM)) %>%
    mutate(tt_class = paste(POI_ID, ind, sep = "_")) %>%
    ungroup() %>%
    select(-ind) %>%
    filter(!COMID %in% POIs$COMID)
  
  if(nrow(tt_splits) > 0) {
    POIs <- POI_creation(select(tt_splits, COMID, tt_class), 
                             filter(flowline, poi == 1), poi_name) %>%
      addType(., POIs, poi_name, nexus = TRUE)
  }
  
  list(POIs = POIs, vpu = vpu)
  
}

#' create final POIs
#' @param POIs POIs from previous step
#' @param flowline flowlines to attach pois to
create_final_pois <- function(POIs, flowline) {
  
  vpu <- POIs$vpu
  POIs <- POIs$POIs

  unCon_POIs <- filter(POIs, COMID %in% filter(flowline, AreaSqKM == 0)$COMID)
  
  xWalk <- NULL
  new_POIs <- NULL
  
  # If any POIs happened to fall on flowlines w/o catchment
  if (nrow(unCon_POIs) > 0){
    # For confluence POIs falling on Flowlines w/o catchments, derive upstream valid flowline,
    poi_fix <- DS_poiFix(POIs, filter(flowline, minNet == 1))
    new_POIs <- st_compatibalize(poi_fix$new_POIs, POIs)
    xWalk <- poi_fix$xWalk
    
    # POIs that didn't need to be moved
    POIs_fixed <- filter(POIs, nexus == TRUE | 
                               !COMID %in% c(poi_fix$xWalk$oldPOI, poi_fix$xWalk$COMID))
    # bind together
    final_POIs <- bind_rows(POIs_fixed, new_POIs) %>%
      mutate(Type_Term = ifelse(nexus == 1, NA, Type_Term)) %>%
      select(-dplyr::any_of(c("ID")))
    
  } else {
    # If no fixes designate as NA
    poi_fix <- NA
    
    POIs$nexus <- as.integer(POIs$nexus)
    
    # if a POI will be a nexus, it can not be terminal.
    final_POIs <- mutate(POIs, Type_Term = ifelse(nexus == 1, NA, Type_Term))
    
  }
 
  list(POIs = final_POIs, xwalk = xWalk, new_POIs = new_POIs, 
       unCon_POIs = unCon_POIs, vpu = vpu) 
}

### used as a cleanup step in POI workflow
#' Moves POI Upstream or downstream if it falls on COMID
#       of flowline with no corresponding catchment
#'  @param POIs_wgeom (sf data.frame) POIs
#'  @param nhdDF  (sf data.frame) valid data frame of NHD flowlines
#' 
#' @return (sf data.frame) data.frame of POIs with new COMID associated
DS_poiFix <- function(POIs_wgeom, nhd){
  nhd <- distinct(nhd)
  POIs <- st_drop_geometry(POIs_wgeom) %>%
    arrange(COMID) %>%
    filter(nexus == FALSE)
  
  # DF of downstream segment
  tocomDF <- select(st_drop_geometry(nhd), COMID, Hydroseq, TotDASqKM,
                    DnHydroseq, WBAREACOMI) %>%
    inner_join(select(st_drop_geometry(nhd), COMID_ds = COMID, Hydroseq, 
                      WBAREACOMI_down = WBAREACOMI, totda_ds = TotDASqKM), 
               by = c("DnHydroseq" = "Hydroseq")) %>%
    inner_join(select(st_drop_geometry(nhd), COMID_us = COMID, DnHydroseq,
                      WBAREACOMI_up = WBAREACOMI, totda_us = TotDASqKM),
               by = c("Hydroseq" = "DnHydroseq"))
  
  # Find segments with POIs where there is no corresponding catchment that are not terminal
  unCon_fl <- filter(nhd, COMID %in% POIs$COMID, AreaSqKM == 0)# & Hydroseq != TerminalPa)
  unCon_POIs <- filter(POIs, COMID %in% unCon_fl$COMID)
  
  # Get specific fixes for waterbody inlets and outlets
  wbout <- filter(unCon_POIs, !is.na(Type_WBOut)) %>%
    inner_join(tocomDF, by = "COMID") %>%
    mutate(nonrefactor = ifelse(WBAREACOMI %in% WBAREACOMI_up, COMID_ds, 0),
           new_POI = COMID_us)
  
  wb_pois <- filter(unCon_POIs, !is.na(Type_WBIn)) %>%
    inner_join(tocomDF, by = "COMID") %>%
    mutate(nonrefactor = ifelse(WBAREACOMI < 0, COMID_ds, 0),
           new_POI = COMID_us) %>%
    rbind(wbout) %>%
    select(-c(nexus, Hydroseq, TotDASqKM, DnHydroseq, WBAREACOMI, COMID_ds,
              WBAREACOMI_down, totda_ds, COMID_us, WBAREACOMI_up, totda_us)) %>%
    rename(COMID = new_POI, oldPOI = COMID)
  
  # The rest can be resolved with drainage are ratio
  unCon_POIs <- filter(unCon_POIs, !COMID %in% wb_pois$oldPOI)
  
  poi_fix <- as.data.frame(do.call("rbind", lapply(unCon_POIs$COMID, movePOI_NA_DA, st_drop_geometry(nhd)))) %>%
    inner_join(POIs, by = c("oldPOI" = "COMID")) %>%
    inner_join(select(st_drop_geometry(nhd), COMID), by = c("oldPOI" = "COMID")) %>%
    select(-c(AreaSqKM, DnHydroseq, nexus, TotDASqKM)) %>%
    distinct() %>%
    bind_rows(wb_pois)
  
  # Fold in new POIs with existing POIs so all the "Type" attribution will carry over
  # using the minimum will ensure correct downstream hydrosequence gets carried over
  poi_orig <- filter(POIs, COMID %in% poi_fix$COMID) %>%
    bind_rows(poi_fix) %>%
    select(-oldPOI)
  
  list_df <- dplyr::group_by(poi_orig, COMID) |>
    group_split()
  
  compact <- function(l) {
    if(nrow(l) == 1) return(as.data.frame(l))
    lapply(names(l), \(x) {
      out <- unique(l[[x]][!is.na(l[[x]])])
      if(!length(out)) NA 
      else if(length(out) == 1) out 
      else {
        cat(paste("duplicate id for", unique(l$COMID),
                  "column", x, 
                  "values", paste(out, collapse = ", "), 
                  "using", out[1]), file = "POI-issues.log")
        out[1]
      }
    }) |> 
      setNames(names(l)) |>
      as.data.frame()
  }
  
  poi_merged <- bind_rows(lapply(list_df, compact))
  
  # # Combine POI information together for redundant pois  
  # poi_merged <- poi_orig %>% 
  #   select(-c(nexus, AreaSqKM, oldPOI, DnHydroseq, TotDASqKM)) %>%
  #   group_by(COMID) %>%
  #   summarise_each(funs(toString(na.omit(.)))) 
  # is.na(poi_merged) <- poi_merged == ""
  
  # Join new POI COMIDs and geometry with the old Type fields
  fin_POIs <- poi_merged %>%
    arrange(COMID) %>%
    bind_cols(get_node(filter(nhd, COMID %in% .$COMID) %>% arrange(COMID), position = "end")) %>%
    st_sf() %>%
    st_compatibalize(., POIs_wgeom)
  
  return (list(xWalk = poi_fix, new_POIs = fin_POIs))
}

### USED in DS_poiFix
#' Move POIs that fall on flowlines with no catchment upstream/downstream
#     to adjacent flowline with most similar total drainage area. Called from 
#     DS_poi_fix function above
#'  @param poi_fix (data.frame) POI data set of COMIDs to be changed
#'  @param nhdDF  (sf data.frame) valid data frame of NHD flowlines
#' 
#' @return (data frame) DF of POIs with new COMID associated
movePOI_NA_DA <- function(poi_fix, nhdDF){
  #print(poi_fix)
  nhdDF <- distinct(nhdDF)
  
  # Closest POI/US/DS
  up_segs <- unique(nhdplusTools::get_UM(nhdDF, poi_fix, sort=T)) 
  seg2fix <- filter(nhdDF, COMID == poi_fix) %>%
    distinct()
  
  # Sorted results and filter out all flowlines w/o catchments
  upstuff <- filter(nhdDF, COMID %in% unlist(up_segs)) %>% 
    arrange(Hydroseq) %>%
    filter(AreaSqKM > 0)
  
  down_segs <- unique(nhdplusTools::get_DM(nhdDF, poi_fix, sort=T))
  downstuff <- filter(nhdDF, COMID %in% unlist(down_segs)) %>% 
    arrange(Hydroseq)%>%
    filter(AreaSqKM > 0)
  
  # combine into one dataframe, select up/downstream seg with least change in total drainage area
  near_FL <- rbind(select(upstuff, COMID, TotDASqKM, AreaSqKM) %>% 
                     slice(1), 
                   select(downstuff, COMID, TotDASqKM, AreaSqKM) %>% 
                     slice(1))
  
  # If 1 or other adjacent flowlines are coupled with a catchment
  if (sum(near_FL$AreaSqKM) > 0){
    new_POI <- near_FL[(which.min(abs(seg2fix$TotDASqKM - near_FL$TotDASqKM))),] #near_FL$COMID
    new_POI$oldPOI <- poi_fix
    new_POI$DnHydroseq <-seg2fix$DnHydroseq
  } else {
    # Remove POI if not catchment associated with flowlines upstream or downstream
    print (poi_fix)
    print ("US and DS flowlines also have no catchment, removing POI")
    new_POI <- NA
  }
  return(new_POI)
}

#' create draft segments
#' @param final_POIs POIs from previous step
#' @param flowline flowlines to attach pois to
create_draft_segments <- function(POIs, flowline) {

  xWalk <- POIs$xwalk
  vpu <- POIs$vpu
  POIs <- POIs$POIs

  if("POI_ID" %in% colnames(flowline)) {
    flowline <- select(flowline, -POI_ID)
  }
  
  # Sort POIs by Levelpath and Hydrosequence in upstream to downstream order
  seg_POIs <-  filter(st_drop_geometry(flowline),  
                      COMID %in% POIs$COMID, 
                      COMID %in% filter(flowline,  minNet == 1)$COMID)

  # derive incremental segments from POIs
  inc_segs <- make_incremental_segments(filter(flowline,  minNet == 1), seg_POIs)
  
  flowline_final <- flowline %>%
    left_join(select(inc_segs, COMID, POI_ID), by = "COMID")
  
  # create and write out final dissolved segments
  nsegments_fin <- segment_creation(flowline_final, xWalk)
  
  flowline_final <- select(flowline_final, -POI_ID) %>%
    left_join(distinct(st_drop_geometry(nsegments_fin$raw_segs), COMID, POI_ID), by = "COMID")
  
  nsegments <- nsegments_fin$diss_segs
  
  list(flowline_final = flowline_final, nsegments = nsegments, vpu = vpu)
  
}

#' Creates finalized segments and routing
#'  @param nhdDF (sf data.frame) valid data frame of NHD flowlines
#'  @param routing_fix  (sf data.frame) any additional routing fixes
#' 
#' @return (sf data.frame) data.frame of segments
segment_creation <- function(nhdDF, routing_fix = NULL){ 
  
  if(!"StartFlag" %in% names(nhdDF)) {
    nhdDF$StartFlag <- ifelse(nhdDF$Hydroseq %in% nhdDF$DnHydroseq, 0, 1)
  }
  
  in_segs <- filter(nhdDF, !is.na(POI_ID))
  
  # If there are routing fixes to account for if a POI with a DA of 0 is moved upsream or downstream
  if (is.data.frame(routing_fix)){
    routing_fix <- routing_fix %>%
      rename(COMID = oldPOI, new_COMID = COMID)
    
    # Above we generated the network using the initial set of POIs; here we crosswalk over the old COMIDs to the new
    nhd_fix <- nhdDF %>%
      left_join(routing_fix %>%
                  select(COMID, new_COMID), by = c("POI_ID" = "COMID")) %>%
      mutate(POI_ID = ifelse(is.na(new_COMID), POI_ID, new_COMID)) %>%
      filter(!POI_ID %in% routing_fix$COMID) %>%
      select(-new_COMID)
    
    in_segs <- filter(nhd_fix, !is.na(POI_ID))
  }
  
  # Dissolve flowlines to aggregated segments
  nsegments <- filter(in_segs, !is.na(POI_ID)) %>%
    group_by(POI_ID) %>%
    #arrange(desc(LevelPathI), desc(Hydroseq)) %>%
    summarize(TotalLength = sum(LENGTHKM),TotalDA = max(TotDASqKM), HW = max(StartFlag),
              do_union = FALSE) %>%
    #st_cast("MULTILINESTRING")  %>%
    inner_join(st_drop_geometry(filter(in_segs, minNet == 1)) %>%
                 select(COMID, Hydroseq, DnHydroseq), by = c("POI_ID" = "COMID"))
  
  # produce a short data frame for populating TO_POI for downstream segment
  to_from <- filter(st_drop_geometry(in_segs)) %>%
    left_join(filter(st_drop_geometry(nhdDF), !is.na(POI_ID)) %>% 
                select(COMID, Hydroseq, POI_ID), by = c("DnHydroseq" = "Hydroseq")) %>%
    select(COMID.x, Hydroseq, DnHydroseq, POI_ID.y) %>%
    rename(To_POI_ID = POI_ID.y) 
  
  # Add To_POI_ID to dissolved segments
  nsegments_fin <- nsegments %>% 
    left_join(select(to_from, COMID.x, To_POI_ID), by = c("POI_ID" = "COMID.x")) %>%
    select(POI_ID, TotalLength, TotalDA, HW, To_POI_ID) 
  
  return(list(diss_segs = nsegments_fin, raw_segs = in_segs))
}

collapse_pois <- function(final_POIs, poi_dar_move, poi_distance_move, flowline) {

  POIs <- final_POIs$POIs
  vpu <- final_POIs$vpu
  
  # number POIs
  POIs <- mutate(POIs, id = row_number(), moved = NA)
  
  moved_pois <- get_moved_pois(POIs, poi_dar_move, poi_distance_move, flowline)
  
  check_dups <- moved_pois$POIs %>%
    group_by(COMID) %>%
    filter(n() > 1) 
  
  dups <- NULL
  
  if(nrow(filter(check_dups, all(c(0,1) %in% nexus))) != nrow(check_dups)){
    print("Multiple POI ids at same geometric location")
    no_dups <- filter(check_dups, all(c(0,1) %in% nexus))
    dups <- filter(check_dups, !id %in% no_dups$id)
    # write_sf(dups, temp_gpkg, dup_pois)
  } else {
    print("All double COMIDs nexus for gage or WB splitting")
  }
  
  list(POIs = POIs, 
       duplicate_pois = dups, 
       pois_collapsed = moved_pois$pois_collapsed,
       vpu = vpu)
}

#'  Collapses POIs together based on criteria
#'  @param pois sf data frame of POIs
#'  @param move_category (character) POI data theme to move
#'  @param DAR (numeric) drainage area threshold to move within
#'  @param dist (numeric) maximum river distance between two points to move within
#'  @param flowline (sf data.frame) nhd flowline
#'  @param keep_category (character) POI data themes to keep static
#' 
#'  @return (sf data.frame, table) dataframe of pois, table of points that have moved
poi_move <- function(pois, move_category, DAR, dist, flowline, keep_category) {
  # filter out features with identical geometry
  
  # Add row_number
  pois_edit <- pois %>%
    mutate(nexus = ifelse(is.na(nexus), 0, nexus))
  
  # Don't consider points already moved
  if("moved" %in% colnames(pois_edit)){
    pois_tomove <- filter(pois_edit, is.na(moved)) # change from poi_edit
    pois_moved_pre <- filter(pois_edit, !is.na(moved))}
  
  # If 'keep' category included
  if(!missing(keep_category)){
    poi2move <- filter(pois_tomove, !is.na(.data[[move_category]]) & nexus == FALSE) %>%
      filter(if_all(!!as.symbol(keep_category), function(x) is.na(x))) %>%
      # Never move these
      filter_at(vars(Type_WBOut, Type_WBIn, Type_Conf, Type_Term), all_vars(is.na(.)))
    
    pois2keep <- filter(pois_tomove, !id %in% poi2move$id) 
    #is.na(.data[[move_category]]) & nexus == FALSE) #%>%
    #filter(if_all(!!as.symbol(keep_category), function(x) is.na(x)))
  } else {
    # POIs to move
    poi2move <- pois_tomove %>%
      filter_at(vars(Type_WBOut, Type_WBIn, Type_Conf, Type_Term), all_vars(is.na(.))) %>%
      filter(nexus == 0) %>%
      filter(!is.na(.data[[move_category]]))
    
    pois2keep <- filter(pois_tomove, !id %in% poi2move$id)
  }
  
  # Get relevant NHD data
  nhd_poi1 <- filter(flowline, COMID %in% pois2keep$COMID)
  nhd_poi2 <- filter(flowline, COMID %in% poi2move$COMID)
  # Ensure they are on same level path
  nhd_poi2 <- filter(nhd_poi2, LevelPathI %in% nhd_poi1$LevelPathI)
  
  # Join NHD data
  pois2keep_nhd <- pois2keep %>% 
    inner_join(select(st_drop_geometry(nhd_poi1), COMID, LevelPathI, Hydroseq,
                      DA_keep = TotDASqKM, Pathlength_keep = Pathlength), by = "COMID") %>%
    rename(COMID_keep = COMID)
  
  # Join NHD data
  pois2move_nhd <- select(poi2move, COMID, !!as.symbol(move_category), id_move = id) %>% 
    inner_join(select(st_drop_geometry(nhd_poi2), COMID, LevelPathI, Hydroseq, TotDASqKM, Pathlength), 
               by = "COMID")
  
  # Candidates to move
  pois2move_cand <-pois2move_nhd %>%
    inner_join(select(st_drop_geometry(pois2keep_nhd), COMID_keep, DA_keep, LevelPathI,
                      Pathlength_keep, id_keep = id, nexus), 
               by = "LevelPathI") %>%
    mutate(river_dist = abs(Pathlength - Pathlength_keep), DAR_poi = abs(DA_keep/TotDASqKM),
           move_dir = ifelse(Pathlength < Pathlength_keep, "Up", "Down")) %>%
    group_by(id_move, move_dir) %>%
    ungroup() %>%
    filter((river_dist < dist) & (DAR_poi > (1 - DAR)) & (DAR_poi < (1 + DAR))) %>%
    select(!!as.symbol(move_category), id_move, COMID, id_keep, COMID_keep, river_dist, DAR_poi, move_dir, nexus) %>%
    st_drop_geometry()
  
  move_distinct <- pois2move_cand %>%
    group_by(id_keep) %>%
    filter(row_number() == 1) %>%
    ungroup() %>%
    distinct(id_move, COMID_move = COMID, id_keep, COMID_keep, river_dist, DAR_poi, move_dir, nexus) %>%
    group_by(id_move) %>%
    slice(which.min(abs(1 - DAR_poi))) 
  
  if(nrow(move_distinct) == 0){
    print("no POIs to move")
    return(pois)
  }
  
  pois2_move <- filter(st_drop_geometry(pois_tomove), id %in% move_distinct$id_move) %>%
    select_if(~sum(!is.na(.)) > 0) %>%
    select(-c(COMID, nexus)) %>%
    inner_join(select(move_distinct, id_move, id_keep), by = c("id" = "id_move"))
  
  move_fields <- colnames(select(pois2_move, -c(id, id_keep)))
  
  if(length(move_fields) == 1){
    pois2_keep <- filter(pois_tomove, id %in% pois2_move$id_keep, !id %in% pois2_move$id) %>%
      inner_join(select(pois2_move, id_move = id, id_keep, 
                        new_val = !!as.symbol(move_category)), by = c("id" = "id_keep")) %>%
      mutate(moved := ifelse(is.na(!!as.symbol(move_category)),
                             id_move, moved),
             !!as.symbol(move_category) := ifelse(is.na(!!as.symbol(move_category)),
                                                  new_val, !!as.symbol(move_category)))
    
    moved_points <- filter(pois2_keep, !is.na(new_val), !is.na(moved)) %>%
      mutate(moved_value = move_category)
  } else {
    for (field in move_fields){
      pois2_keep <- filter(pois_tomove, id %in% pois2_move$id_keep, !id %in% pois2_move$id) %>%
        inner_join(select(pois2_move, id_move = id, id_keep, new_val = !!as.symbol(field)), 
                   by = c("id" = "id_keep")) %>%
        mutate(moved := ifelse(is.na(!!as.symbol(field)),
                               id_move, moved),
               !!as.symbol(field) := ifelse(is.na(!!as.symbol(field)),
                                            new_val, !!as.symbol(field)))
      
      pois_moved <- filter(pois2_keep, !is.na(new_val), !is.na(moved)) %>%
        mutate(moved_value = field)
      
      if(!exists("moved_points")){
        moved_points <- pois_moved
      } else {
        moved_points <- rbind(moved_points, pois_moved)
      }
    }
  }
  
  
  pois_final <- data.table::rbindlist(list(filter(pois_edit, !id %in% c(moved_points$id_move, pois2_keep$id)),
                                           select(pois2_keep, -c(new_val, id_move, new_val))), fill = TRUE) %>%
    st_as_sf()
  
  return(list(final_pois = pois_final, moved_points = moved_points))
  
}


#' get moved pois
#' @param final_POIs finalized POIs that can be reconciled
#' @param poi_dar_move POI drainage area ratio where moves are allowed
#' @param poi_distance_move POI distance where moves are allowed
#' @param flowline flowline to use for evaluation of poi moves 
get_moved_pois <- function(final_POIs, poi_dar_move, poi_distance_move, flowline) {
  
  # Move HUC12 to other POIs
  moved_pois <- poi_move(final_POIs, "Type_HUC12", poi_dar_move, 
                         poi_distance_move, flowline) 
  
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois$moved_points %>%
      mutate(move_type = "huc12 to other")
  } else {
    final_POIs <- moved_POIs
  }
  
  # Gages to confluences, terminals
  moved_pois <- poi_move(final_POIs, "Type_Gages", poi_dar_move, 
                         poi_distance_move, flowline, c("Type_Conf", "Type_Term"))
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "gages to conf"))
  } else {
    final_POIs <- moved_POIs
  }
  
  # Gages to waterbody inlets
  moved_pois <- poi_move(final_POIs, "Type_Gages", poi_dar_move, 
                         poi_distance_move, flowline, c("Type_WBIn", "Type_WBOut")) 
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "gages to wbin"))
  } else {
    final_POIs <- moved_pois
  }
  
  # Waterbody inlet to confluence
  moved_pois <- poi_move(final_POIs, "Type_WBIn", poi_dar_move/2, 
                         poi_distance_move*0.4, flowline, "Type_Conf")
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "gages to term"))
  } else {
    final_POIs <- moved_pois
  }
  
  # # Waterbody inlet to confluence
  # moved_pois <- poi_move(final_POIs, "Type_WBOut", poi_dar_move/2, 
  #                        poi_distance_move*0.4, flowline, "Type_WBIn")
  # if(!is.data.frame(moved_pois)){
  #   final_POIs <- moved_pois$final_pois
  #   moved_pois_table <- moved_pois_table %>%
  #     rbind(moved_pois$moved_points %>%
  #             mutate(move_type = "gages to term"))
  # } else {
  #   final_POIs <- moved_pois
  # }
  
  # Waterbody inlet to confluence
  # TODO: verify that this is OK Type_HUC12 was here in move_category?!?
  moved_pois <- poi_move(final_POIs, "Type_WBIn", poi_dar_move/2, 
                         poi_distance_move*0.4, flowline, "Type_Conf")
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "gages to term"))
  } else {
    final_POIs <- moved_pois
  }
  
  # NID to waterbody outlet
  moved_pois <- poi_move(final_POIs, "Type_hilarri", poi_dar_move/2, 
                         poi_distance_move * 0.4, flowline, c("Type_WBOut", "Type_TE"))
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "nid to wb_out"))
  } else {
    final_POIs <- moved_pois
  }
  
  # NID to waterbody outlet
  moved_pois <- poi_move(final_POIs, "Type_DA", poi_dar_move, 
                         poi_distance_move, flowline)
  if(!is.data.frame(moved_pois)){
    final_POIs <- moved_pois$final_pois
    moved_pois_table <- moved_pois_table %>%
      rbind(moved_pois$moved_points %>%
              mutate(move_type = "nid to wb_out"))
  } else {
    final_POIs <- moved_pois
  }
  
  if("Type_elev" %in% names(final_POIs)){
    # NID to waterbody outlet
    moved_pois <- poi_move(final_POIs, "Type_elev", poi_dar_move, 
                           poi_distance_move, flowline)
    if(!is.data.frame(moved_pois)){
      final_POIs <- moved_pois$final_pois
      moved_pois_table <- moved_pois_table %>%
        rbind(moved_pois$moved_points %>%
                mutate(move_type = "nid to wb_out"))
    } else {
      final_POIs <- moved_pois
    }
  }
  
  list(POIs = final_POIs, pois_collapsed = moved_pois_table)
  
} 

create_poi_lookup <- function(POIs, events, full_cats, 
                              flowline) {

  vpu <- POIs$vpu
  
  pois_collapsed <- POIs$pois_collapsed
  
  # Final POI layer
  POIs <- POIs$POIs %>%
    mutate(identifier = row_number())
  
  # Unique POI geometry
  final_POI_geom <- POIs %>%
    select(identifier) %>%
    cbind(st_coordinates(.)) %>%
    group_by(X, Y) %>%
    mutate(geom_id = cur_group_id()) %>%
    ungroup()
  
  final_POIs_table <- POIs %>%
    inner_join(select(st_drop_geometry(final_POI_geom), -X, -Y), by = "identifier")  %>%
    select(-identifier) 
  
  # POI data theme table
  pois_data_orig <- reshape2::melt(st_drop_geometry(select(final_POIs_table,
                                                           -c(nexus, nonrefactor, id, moved))),
                                   id.vars = c("COMID", "geom_id")) %>%
    filter(!is.na(value)) %>%
    group_by(COMID, geom_id) %>%
    mutate(identifier = cur_group_id()) %>%
    rename(hy_id = COMID, poi_id = identifier, hl_reference = variable, hl_link = value) %>%
    distinct() 
  
  if(!is.null(pois_collapsed)){
    pois_data_moved <- select(st_drop_geometry(pois_collapsed), 
                              hy_id = COMID, hl_link = new_val, hl_reference = moved_value) %>%
      inner_join(distinct(pois_data_orig, hy_id, geom_id, poi_id), by = "hy_id") 
    
    pois_data <- data.table::rbindlist(list(pois_data_moved, pois_data_orig), use.names = TRUE) %>%
      filter(!hl_reference %in% c("id", "moved"))
  } else {
    pois_data <- pois_data_orig
  }
  
  # POI Geometry table
  poi_geometry <- select(final_POIs_table, hy_id = COMID, geom_id) %>%
    inner_join(distinct(pois_data, hy_id, geom_id, poi_id),
               by = c("hy_id" = "hy_id", "geom_id" = "geom_id")) %>%
    distinct() %>%
    st_as_sf()
  
  # write_sf(pois_data, nav_gpkg, poi_data_table)
  # write_sf(poi_geometry, nav_gpkg, poi_geometry_table)
  
  poi_geom_xy <- cbind(poi_geometry, st_coordinates(poi_geometry)) %>%
    st_drop_geometry()
  
  event_table <- NULL
  
  if(!is.null(events)) {
    events_data <- events %>%
      arrange(COMID) %>%
      cbind(st_coordinates(.)) %>%
      st_drop_geometry() %>%
      group_by(COMID, REACHCODE, REACH_meas) %>%
      mutate(event_id = cur_group_id()) %>%
      rename(hy_id = COMID) %>%
      ungroup()
    
    event_table <- select(events_data, -c(nexus, X, Y))
  }
  
  nexi <- filter(final_POIs_table, nexus == 1) %>%
    cbind(st_coordinates(.)) %>%
    select(hy_id = COMID, X, Y) %>%
    inner_join(poi_geom_xy, by = c("hy_id" = "hy_id", "X" = "X", "Y" = "Y")) %>%
    inner_join(events_data, by = c("hy_id" = "hy_id", "X" = "X", "Y" = "Y"), multiple = "all") %>%
    select(hy_id, REACHCODE, REACH_meas, event_id, poi_id) %>%
    group_by(hy_id, REACHCODE) %>%
    filter(REACH_meas == min(REACH_meas)) %>%
    ungroup()
  #distinct(hy_id, REACHCODE, REACH_meas, event_id, poi_id)
  
  event_geometry_table <- nexi
  
  # write_sf(, nav_gpkg, event_table)
  # write_sf(nexi, nav_gpkg, event_geometry_table)
  
  #  Load data
    lookup <- dplyr::select(sf::st_drop_geometry(flowline),
                            NHDPlusV2_COMID = COMID,
                            realized_catchmentID = COMID,
                            mainstem = LevelPathI) %>%
      dplyr::mutate(realized_catchmentID = ifelse(realized_catchmentID %in% full_cats$FEATUREID,
                                                  realized_catchmentID, NA)) %>%
      left_join(select(st_drop_geometry(poi_geometry), hy_id, poi_geom_id = geom_id), 
                by = c("NHDPlusV2_COMID" = "hy_id"))
    
    # sf::write_sf(lookup, nav_gpkg, lookup_table_refactor)
  
  list(lookup = lookup, event_table = event_table, 
       event_geometry_table = event_geometry_table, 
       pois_data = pois_data, poi_geometry = poi_geometry, vpu = vpu)
  
}